# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import oslotest.base
from oslo_log import log as logging

from designate import exceptions
from designate import objects

LOG = logging.getLogger(__name__)


class RRDataSPFTest(oslotest.base.BaseTestCase):
    def test_reject_non_quoted_spaces(self):
        record = objects.SPF(data='foo bar')
        self.assertRaisesRegex(
            exceptions.InvalidObject,
            'Provided object does not match schema',
            record.validate
        )

    def test_reject_non_escaped_quotes(self):
        record = objects.SPF(data='foo"bar')
        self.assertRaisesRegex(
            exceptions.InvalidObject,
            'Provided object does not match schema',
            record.validate
        )



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Author: Alin Balutoiu <abalutoiu@cloudbasesolutions.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_config import cfg

MSDNS_GROUP = cfg.OptGroup(
    name='backend:agent:msdns',
    title="Configuration for Microsoft DNS Server"
)
MSDNS_OPTS = [
]


def register_opts(conf):
    conf.register_group(MSDNS_GROUP)
    conf.register_opts(MSDNS_OPTS, group=MSDNS_GROUP)


def list_opts():
    return {
        MSDNS_GROUP: MSDNS_OPTS,
    }



# Copyright 2015 Dyn Inc.
#
# Author: Yasha Bubnov <ybubnov@dyn.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import itertools

import dns.rdata
import dns.rdatatype
import dns.rdataclass
import six
from oslo_config import cfg
from oslo_concurrency import lockutils
from oslo_log import log as logging

from designate.backend.agent_backend import base
from designate import exceptions
from designate import utils

CFG_GROUP_NAME = 'backend:agent:denominator'
LOG = logging.getLogger(__name__)


class Denominator(object):

    def __init__(self, config):
        super(Denominator, self).__init__()
        self.config = config

    def update_record(self, zone, **kwargs):
        return self._execute(['record', '-z', zone, 'replace'], kwargs)

    def create_record(self, zone, **kwargs):
        return self._execute(['record', '-z', zone, 'add'], kwargs)

    def delete_record(self, zone, **kwargs):
        return self._execute(['record', '-z', zone, 'delete'], kwargs)

    def get_record(self, zone, **kwargs):
        return self._execute(['record', '-z', zone, 'get'], kwargs)

    def get_records(self, zone, **kwargs):
        return self._execute(['record', '-z', zone, 'list'], kwargs)

    def create_zone(self, **kwargs):
        return self._execute(['zone', 'add'], kwargs)

    def update_zone(self, **kwargs):
        return self._execute(['zone', 'update'], kwargs)

    def delete_zone(self, **kwargs):
        return self._execute(['zone', 'delete'], kwargs)

    def _params(self, **kwargs):
        params = [('--%s' % k, str(v)) for k, v in kwargs.items()]
        return list(itertools.chain(*params))

    def _base(self):
        call = ['denominator', '-q', '-n', self.config.name]

        # NOTE: When path to denominator configuration file is omitted,
        #       ~/.denominatorconfig file will be used by default.
        if self.config.config_file:
            call.extend(['-C', self.config.config_file])
        return call

    def _execute(self, op, kwargs):
        try:
            call = self._base() + op + self._params(**kwargs)
            LOG.debug(('Executing Denominator call: %s' % ' '.join(call)))

            stdout, _ = utils.execute(*call)
            return stdout
        except utils.processutils.ProcessExecutionError as e:
            LOG.debug('Denominator call failure: %s' % e)
            raise exceptions.DesignateException(e)


class DenominatorBackend(base.AgentBackend):
    __plugin_name__ = 'denominator'

    __backend_status__ = 'untested'

    def __init__(self, agent_service):
        super(DenominatorBackend, self).__init__(agent_service)
        self.denominator = Denominator(
            cfg.CONF[CFG_GROUP_NAME])

    def start(self):
        LOG.info("Started Denominator backend")

    def stop(self):
        LOG.info("Stopped Denominator backend")

    def find_zone_serial(self, zone_name):
        LOG.debug("Finding %s", zone_name)

        zone_name = zone_name.rstrip('.')
        output = self.denominator.get_record(
            zone=zone_name,
            type='SOA',
            name=zone_name)
        try:
            text = ' '.join(output.split()[3:])
            rdata = dns.rdata.from_text(dns.rdataclass.IN,
                                        dns.rdatatype.SOA,
                                        text)
        except Exception:
            return None
        return rdata.serial

    def create_zone(self, zone):
        LOG.debug("Creating %s", zone.origin.to_text())
        zone_name = zone.origin.to_text(omit_final_dot=True)
        if six.PY3 and isinstance(zone_name, bytes):
            zone_name = zone_name.decode('utf-8')

        # Use SOA TTL as zone default TTL
        soa_record = zone.find_rrset(zone.origin, dns.rdatatype.SOA)
        rname = soa_record.items[0].rname.derelativize(origin=zone.origin)

        # Lock zone to prevent concurrent changes.
        with self._sync_zone(zone.origin):
            # NOTE: If zone already exists, denominator will update it with
            #       new values, in other a duplicate zone will be created if
            #       provider supports such functionality.
            self.denominator.create_zone(
                name=zone_name,
                ttl=soa_record.ttl,
                email=rname)

            # Add records one by one.
            for name, ttl, rtype, data in self._iterate_records(zone):
                # Some providers do not support creation of SOA record.
                rdatatype = dns.rdatatype.from_text(rtype)
                if rdatatype == dns.rdatatype.SOA:
                    continue

                self.denominator.create_record(
                    zone=zone_name,
                    name=name,
                    type=rtype,
                    ttl=ttl,
                    data=data)

    def update_zone(self, zone):
        LOG.debug("Updating %s", zone.origin)
        zone_name = zone.origin.to_text(omit_final_dot=True)
        if six.PY3 and isinstance(zone_name, bytes):
            zone_name = zone_name.decode('utf-8')

        soa_record = zone.find_rrset(zone.origin, dns.rdatatype.SOA)
        rname = soa_record.items[0].rname.derelativize(origin=zone.origin)

        with self._sync_zone(zone.origin):
            # Update zone with a new parameters
            self.denominator.update_zone(
                id=zone_name,
                ttl=soa_record.ttl,
                email=rname)

            # Fetch records to create a differential update of a zone.
            output = self.denominator.get_records(zone_name)
            subzones = dict()

            # subzones dict will contain names of subzones without
            # trailing dot.
            for raw in output.splitlines():
                data = raw.split()
                name, rtype = data[0], data[1]

                rtypes = subzones.get(name, set())
                rtypes.add(rtype)
                subzones[name] = rtypes

            for name, ttl, rtype, data in self._iterate_records(zone):
                record_action = self.denominator.create_record

                if name in subzones and rtype in subzones[name]:
                    # When RR set already exists, replace it with a new one.
                    rdatatype = dns.rdatatype.from_text(rtype)
                    record_action = self.denominator.update_record

                    # So next call will ADD a new record to record set
                    # instead of replacing of the existing one.
                    subzones[name].remove(rtype)

                    # NOTE: DynECT does not support deleting of the SOA
                    #       record. Skip updating of the SOA record.
                    if rdatatype == dns.rdatatype.SOA:
                        continue

                record_action(zone=zone_name,
                              name=name,
                              type=rtype,
                              ttl=ttl,
                              data=data)

            # Remaining records should be deleted
            for name, types in subzones.items():
                for rtype in types:
                    self.denominator.delete_record(
                        zone=zone_name, id=name, type=rtype)

    def delete_zone(self, zone_name):
        LOG.debug('Delete Zone: %s' % zone_name)

        with self._sync_zone(zone_name):
            self.denominator.delete_zone(id=zone_name)

    def _sync_zone(self, zone_name):
        LOG.debug('Synchronising zone: %s' % zone_name)
        return lockutils.lock('denominator-%s' % zone_name)

    def _iterate_records(self, zone):
        for rname, ttl, rdata in zone.iterate_rdatas():
            name = rname.derelativize(origin=zone.origin)
            name = name.to_text(omit_final_dot=True)
            if six.PY3 and isinstance(name, bytes):
                name = name.decode('utf-8')

            data = rdata.to_text(origin=zone.origin, relativize=False)
            yield name, ttl, dns.rdatatype.to_text(rdata.rdtype), data



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import patch
from unittest import mock

from oslo_config import cfg
import oslo_messaging as messaging

from designate import exceptions
from designate import objects
from designate.central import service as central_service
from designate.mdns import rpcapi as mdns_api
from designate.tests.test_api.test_v2 import ApiV2TestCase


class ApiV2ZonesTest(ApiV2TestCase):
    def setUp(self):
        super(ApiV2ZonesTest, self).setUp()

        # Create the default TLDs
        self.create_default_tlds()

    def test_create_zone(self):
        # Create a zone
        fixture = self.get_zone_fixture(fixture=0)
        response = self.client.post_json('/zones/', fixture)
        # Check the headers are what we expect
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual('PENDING', response.json['status'])
        self.assertEqual('PRIMARY', response.json['type'])
        self.assertEqual([], response.json['masters'])
        self.assertIsNone(response.json['updated_at'])

        for k in fixture:
            self.assertEqual(fixture[k], response.json[k])

    def test_create_zone_no_type(self):
        # Create a zone
        fixture = self.get_zone_fixture(fixture=0)
        del fixture['type']

        response = self.client.post_json('/zones/', fixture)

        # Check the headers are what we expect
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual('PENDING', response.json['status'])
        self.assertEqual('PRIMARY', response.json['type'])
        self.assertEqual([], response.json['masters'])
        self.assertIsNone(response.json['updated_at'])

        for k in fixture:
            self.assertEqual(fixture[k], response.json[k])

    def test_create_zone_validation(self):
        # NOTE: The schemas should be tested separately to the API. So we
        #       don't need to test every variation via the API itself.
        # Fetch a fixture
        fixture = self.get_zone_fixture(fixture=0)

        # Add a junk field to the body
        fixture['junk'] = 'Junk Field'

        # Ensure it fails with a 400
        body = fixture

        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_email_too_long(self):
        fixture = self.get_zone_fixture(fixture=0)
        fixture.update({'email': 'a' * 255 + '@abc.com'})
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_invalid_email(self):
        invalid_emails = [
            'org',
            'example.org',
            'bla.example.org',
            'org.',
            'example.org.',
            'bla.example.org.',
        ]
        fixture = self.get_zone_fixture(fixture=0)
        for email in invalid_emails:
            fixture.update({'email': email})
            body = fixture
            self._assert_exception('invalid_object', 400,
                                   self.client.post_json,
                                   '/zones', body)

    def test_create_zone_email_missing(self):
        fixture = self.get_zone_fixture(fixture=0)
        del fixture['email']
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_ttl_less_than_zero(self):
        fixture = self.get_zone_fixture(fixture=0)
        fixture['ttl'] = -1
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_ttl_is_zero(self):
        fixture = self.get_zone_fixture(fixture=0)
        fixture['ttl'] = 0
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_ttl_is_greater_than_max(self):
        fixture = self.get_zone_fixture(fixture=0)
        fixture['ttl'] = 2174483648
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_ttl_is_invalid(self):
        fixture = self.get_zone_fixture(fixture=0)
        fixture['ttl'] = "!@?>"
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_ttl_is_not_required_field(self):
        fixture = self.get_zone_fixture(fixture=0)
        body = fixture
        response = self.client.post_json('/zones', body)
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)

    def test_create_zone_description_too_long(self):
        fixture = self.get_zone_fixture(fixture=0)
        fixture['description'] = "a" * 161
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_name_is_missing(self):
        fixture = self.get_zone_fixture(fixture=0)
        del fixture['name']
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_name_too_long(self):
        fixture = self.get_zone_fixture(fixture=0)
        fixture['name'] = 'x' * 255 + ".com"
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_body_validation(self):
        fixture = self.get_zone_fixture(fixture=0)
        # Add id to the body
        fixture['id'] = '2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'
        # Ensure it fails with a 400
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

        fixture = self.get_zone_fixture(fixture=0)
        # Add created_at to the body
        fixture['created_at'] = '2014-03-12T19:07:53.000000'
        # Ensure it fails with a 400
        body = fixture
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', body)

    def test_create_zone_invalid_name(self):
        # Try to create a zone with an invalid name
        fixture = self.get_zone_fixture(fixture=-1)

        # Ensure it fails with a 400
        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones', fixture)

    @patch.object(central_service.Service, 'create_zone',
                  side_effect=messaging.MessagingTimeout())
    def test_create_zone_timeout(self, _):
        fixture = self.get_zone_fixture(fixture=0)

        body = fixture

        self._assert_exception('timeout', 504, self.client.post_json,
                               '/zones/', body)

    @patch.object(central_service.Service, 'create_zone',
                  side_effect=exceptions.DuplicateZone())
    def test_create_zone_duplicate(self, _):
        fixture = self.get_zone_fixture(fixture=0)

        body = fixture

        self._assert_exception('duplicate_zone', 409, self.client.post_json,
                               '/zones/', body)

    def test_create_zone_missing_content_type(self):
        self._assert_exception('unsupported_content_type', 415,
                               self.client.post, '/zones')

    def test_create_zone_bad_content_type(self):
        self._assert_exception(
            'unsupported_content_type', 415, self.client.post, '/zones',
            headers={'Content-type': 'test/goat'})

    def test_zone_invalid_url(self):
        url = '/zones/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980/invalid'
        self._assert_exception('not_found', 404, self.client.get, url,
                               headers={'Accept': 'application/json'})
        self._assert_exception('not_found', 404, self.client.patch_json, url)
        self._assert_exception('not_found', 404, self.client.delete, url)

        # Pecan returns a 405 for post
        response = self.client.post(url, status=405)
        self.assertEqual(405, response.status_int)

    def test_get_zones(self):
        response = self.client.get('/zones/')

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('zones', response.json)
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # We should start with 0 zones
        self.assertEqual(0, len(response.json['zones']))

        # We should start with 0 zones
        self.assertEqual(0, len(response.json['zones']))

        data = [self.create_zone(name='x-%s.com.' % i)
                for i in 'abcdefghij']
        self._assert_paging(data, '/zones', key='zones')

        self._assert_invalid_paging(data, '/zones', key='zones')

    @patch.object(central_service.Service, 'find_zones',
                  side_effect=messaging.MessagingTimeout())
    def test_get_zones_timeout(self, _):
        self._assert_exception('timeout', 504, self.client.get, '/zones/')

    def test_get_zone(self):
        # Create a zone
        zone = self.create_zone()

        response = self.client.get('/zones/%s' % zone['id'],
                                   headers=[('Accept', 'application/json')])

        # Check the headers are what we expect
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual('PENDING', response.json['status'])
        self.assertIsNone(response.json['updated_at'])
        self.assertEqual(zone['name'], response.json['name'])
        self.assertEqual(zone['email'], response.json['email'])

    def test_get_zone_invalid_id(self):
        self._assert_invalid_uuid(self.client.get, '/zones/%s')

    @patch.object(central_service.Service, 'get_zone',
                  side_effect=messaging.MessagingTimeout())
    def test_get_zone_timeout(self, _):
        url = '/zones/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'
        self._assert_exception('timeout', 504, self.client.get, url,
                               headers={'Accept': 'application/json'})

    @patch.object(central_service.Service, 'get_zone',
                  side_effect=exceptions.ZoneNotFound())
    def test_get_zone_missing(self, _):
        url = '/zones/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'
        self._assert_exception('zone_not_found', 404, self.client.get, url,
                               headers={'Accept': 'application/json'})

    def test_get_zone_bad_accept(self):
        url = '/zones/6e2146f3-87bc-4f47-adc5-4df0a5c78218'

        self.client.get(url, headers={'Accept': 'test/goat'}, status=406)

    def test_update_zone(self):
        # Create a zone
        zone = self.create_zone()

        # Prepare an update body
        body = {'email': 'prefix-%s' % zone['email']}

        response = self.client.patch_json('/zones/%s' % zone['id'], body,
                                          status=202)

        # Check the headers are what we expect
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])
        self.assertIn('status', response.json)

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIsNotNone(response.json['updated_at'])
        self.assertEqual('prefix-%s' % zone['email'],
                         response.json['email'])

    def test_update_zone_invalid_id(self):
        self._assert_invalid_uuid(self.client.patch_json, '/zones/%s')

    def test_update_zone_validation(self):
        # NOTE: The schemas should be tested separatly to the API. So we
        #       don't need to test every variation via the API itself.
        # Create a zone
        zone = self.create_zone()

        # Prepare an update body with junk in the body
        body = {'email': 'prefix-%s' % zone['email'],
                'junk': 'Junk Field'}

        url = '/zones/%s' % zone['id']

        # Ensure it fails with a 400
        self._assert_exception('invalid_object', 400, self.client.patch_json,
                               url, body)

        # Prepare an update body with negative ttl in the body
        body = {'email': 'prefix-%s' % zone['email'],
                'ttl': -20}

        # Ensure it fails with a 400
        self._assert_exception('invalid_object', 400, self.client.patch_json,
                               url, body)

        # Prepare an update body with ttl > maximum (2147483647) in the body
        body = {'email': 'prefix-%s' % zone['email'],
                'ttl': 2147483648}

        # Ensure it fails with a 400
        self._assert_exception('invalid_object', 400, self.client.patch_json,
                               url, body)

    @patch.object(central_service.Service, 'get_zone',
                  side_effect=exceptions.DuplicateZone())
    def test_update_zone_duplicate(self, _):
        # Prepare an update body
        body = {'email': 'example@example.org'}

        url = '/zones/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        # Ensure it fails with a 409
        self._assert_exception('duplicate_zone', 409, self.client.patch_json,
                               url, body)

    @patch.object(central_service.Service, 'get_zone',
                  side_effect=messaging.MessagingTimeout())
    def test_update_zone_timeout(self, _):
        # Prepare an update body
        body = {'email': 'example@example.org'}

        url = '/zones/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        # Ensure it fails with a 504
        self._assert_exception('timeout', 504, self.client.patch_json,
                               url, body)

    @patch.object(central_service.Service, 'get_zone',
                  side_effect=exceptions.ZoneNotFound())
    def test_update_zone_missing(self, _):
        # Prepare an update body
        body = {'email': 'example@example.org'}

        url = '/zones/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        # Ensure it fails with a 404
        self._assert_exception('zone_not_found', 404, self.client.patch_json,
                               url, body)

    def test_delete_zone(self):
        zone = self.create_zone()

        response = self.client.delete('/zones/%s' % zone['id'], status=202)

        # Check the headers are what we expect
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual('DELETE', response.json['action'])
        self.assertEqual('PENDING', response.json['status'])

        # The deleted zone should still be listed
        zones = self.client.get('/zones/')
        self.assertEqual(1, len(zones.json['zones']))

    def test_delete_zone_invalid_id(self):
        self._assert_invalid_uuid(self.client.delete, '/zones/%s')

    @patch.object(central_service.Service, 'delete_zone',
                  side_effect=messaging.MessagingTimeout())
    def test_delete_zone_timeout(self, _):
        url = '/zones/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        self._assert_exception('timeout', 504, self.client.delete, url)

    @patch.object(central_service.Service, 'delete_zone',
                  side_effect=exceptions.ZoneNotFound())
    def test_delete_zone_missing(self, _):
        url = '/zones/2fdadfb1-cf96-4259-ac6b-bb7b6d2ff980'

        self._assert_exception('zone_not_found', 404, self.client.delete,
                               url)

    def test_post_abandon_zone(self):
        zone = self.create_zone()
        url = '/zones/%s/tasks/abandon' % zone.id

        # Ensure that we get permission denied
        self._assert_exception('forbidden', 403, self.client.post_json, url)

        # Ensure that abandon zone succeeds with the right policy
        self.policy({'abandon_zone': '@'})
        response = self.client.post_json(url)
        self.assertEqual(204, response.status_int)

    def test_get_abandon_zone(self):
        zone = self.create_zone()
        url = '/zones/%s/tasks/abandon' % zone.id
        self._assert_exception('method_not_allowed', 405, self.client.get, url)

    def test_get_invalid_abandon(self):
        # This is an invalid endpoint - should return 404
        url = '/zones/tasks/abandon'
        self._assert_exception('not_found', 404, self.client.get, url)

    def test_get_zone_tasks(self):
        # This is an invalid endpoint - should return 404
        zone = self.create_zone()
        url = '/zones/%s/tasks' % zone.id
        self._assert_exception('not_found', 404, self.client.get, url)

    def test_create_secondary(self):
        # Create a zone
        fixture = self.get_zone_fixture('SECONDARY', 0)
        fixture['masters'] = ["10.0.0.1"]

        response = self.client.post_json('/zones/', fixture)

        # Check the headers are what we expect
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIn('created_at', response.json)
        self.assertEqual('PENDING', response.json['status'])
        self.assertEqual(cfg.CONF['service:central'].managed_resource_email,
                         response.json['email'])

        self.assertIsNone(response.json['updated_at'])
        # Zone is not transferred yet
        self.assertIsNone(response.json['transferred_at'])
        # Serial defaults to 1
        self.assertEqual(response.json['serial'], 1)

        for k in fixture:
            self.assertEqual(fixture[k], response.json[k])

    def test_create_secondary_no_masters(self):
        # Create a zone
        fixture = self.get_zone_fixture('SECONDARY', 0)

        self._assert_exception('invalid_object', 400, self.client.post_json,
                               '/zones/', fixture)

    def test_update_secondary(self):
        # Create a zone
        zone = objects.Zone(
            name='example.com.',
            type='SECONDARY',
            masters=objects.ZoneMasterList.from_list([
                {'host': '1.0.0.0', 'port': 69},
                {'host': '2.0.0.0', 'port': 69}
            ])
        )
        zone.email = cfg.CONF['service:central'].managed_resource_email

        # Create a zone
        zone = self.central_service.create_zone(self.admin_context, zone)

        masters = ['10.0.0.1', '10.0.0.2']

        # Prepare an update body
        body = {'masters': masters}

        response = self.client.patch_json('/zones/%s' % zone['id'], body,
                                          status=202)

        # Check the headers are what we expect
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)

        # Check the body structure is what we expect
        self.assertIn('links', response.json)
        self.assertIn('self', response.json['links'])
        self.assertIn('status', response.json)

        # Check the values returned are what we expect
        self.assertIn('id', response.json)
        self.assertIsNotNone(response.json['updated_at'])
        self.assertEqual(masters, response.json['masters'])
        self.assertEqual(1, response.json['serial'])

    def test_xfr_request(self):
        # Create a zone
        fixture = self.get_zone_fixture('SECONDARY', 0)
        fixture['email'] = cfg.CONF['service:central'].managed_resource_email
        fixture['masters'] = [{"host": "10.0.0.10", "port": 53}]

        # Create a zone
        zone = self.create_zone(**fixture)

        mdns = mock.Mock()
        with mock.patch.object(mdns_api.MdnsAPI, 'get_instance') as get_mdns:
            get_mdns.return_value = mdns
            mdns.get_serial_number.return_value = ('SUCCESS', 10, 1, )

            response = self.client.post_json(
                '/zones/%s/tasks/xfr' % zone['id'],
                None, status=202)

        self.assertTrue(mdns.perform_zone_xfr.called)

        # Check the headers are what we expect
        self.assertEqual(202, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertEqual(b'""', response.body)

    def test_invalid_xfr_request(self):
        # Create a zone
        zone = self.create_zone()

        response = self.client.post_json(
            '/zones/%s/tasks/xfr' % zone['id'],
            None, status=400)

        # Check the headers are what we expect
        self.assertEqual(400, response.status_int)
        self.assertEqual('application/json', response.content_type)

    def test_update_secondary_email_invalid_object(self):
        # Create a zone
        fixture = self.get_zone_fixture('SECONDARY', 0)
        fixture['email'] = cfg.CONF['service:central'].managed_resource_email

        # Create a zone
        zone = self.create_zone(**fixture)

        body = {'email': 'foo@bar.io'}

        self._assert_exception('invalid_object', 400, self.client.patch_json,
                               '/zones/%s' % zone['id'], body)

    # Metadata tests
    def test_metadata_exists(self):
        response = self.client.get('/zones/')

        # Make sure the fields exist
        self.assertIn('metadata', response.json)
        self.assertIn('total_count', response.json['metadata'])

    def test_total_count(self):
        response = self.client.get('/zones/')

        # There are no zones by default
        self.assertEqual(0, response.json['metadata']['total_count'])

        # Create a zone
        fixture = self.get_zone_fixture(fixture=0)
        response = self.client.post_json('/zones/', fixture)

        response = self.client.get('/zones/')

        # Make sure total_count picked it up
        self.assertEqual(1, response.json['metadata']['total_count'])

    def test_total_count_pagination(self):
        # Create two zones
        fixture = self.get_zone_fixture(fixture=0)
        response = self.client.post_json('/zones/', fixture)

        fixture = self.get_zone_fixture(fixture=1)
        response = self.client.post_json('/zones/', fixture)

        # Paginate so that there is only one zone returned
        response = self.client.get('/zones?limit=1')

        self.assertEqual(1, len(response.json['zones']))

        # The total_count should know there are two
        self.assertEqual(2, response.json['metadata']['total_count'])

    def test_no_update_deleting(self):
        # Create a zone
        zone = self.create_zone()

        # Prepare an update body
        body = {'zone': {'email': 'prefix-%s' % zone['email']}}

        self.client.delete('/zones/%s' % zone['id'], status=202)
        self._assert_exception('bad_request', 400, self.client.patch_json,
                               '/zones/%s' % zone['id'], body)

    def test_get_nameservers(self):
        # Create a zone
        zone = self.create_zone()

        # Prepare an update body

        response = self.client.get('/zones/%s/nameservers' % zone['id'],
                                   headers=[('Accept', 'application/json')])

        self.assertIn('nameservers', response.json)
        self.assertEqual(1, len(response.json['nameservers']))
        self.assertIn('hostname', response.json['nameservers'][0])
        self.assertIn('priority', response.json['nameservers'][0])

    def test_get_zones_filter(self):
        # Add zones for testing
        fixtures = [
            self.get_zone_fixture(
                'PRIMARY', fixture=0, values={
                    'ttl': 3600,
                    'description': 'test1'
                }
            ),
            self.get_zone_fixture(
                'PRIMARY', fixture=1, values={
                    'ttl': 4000,
                    'description': 'test2'
                }
            )
        ]

        for fixture in fixtures:
            response = self.client.post_json('/zones/', fixture)

        get_urls = [
            # Filter by Type
            '/zones?type=%s' % fixtures[0]['type'],

            # Filter by Name
            '/zones?name=%s' % fixtures[0]['name'],

            # Filter by Email
            '/zones?email=example*',
            '/zones?email=%s' % fixtures[1]['email'],

            # Filter by TTL
            '/zones?ttl=3600',

            # Filter by Description
            '/zones?description=test1',
            '/zones?description=test*'
        ]

        correct_results = [2, 1, 2, 1, 1, 1, 2]

        for get_url, correct_result in zip(get_urls, correct_results):

            response = self.client.get(get_url)

            # Check the headers are what we expect
            self.assertEqual(200, response.status_int)
            self.assertEqual('application/json', response.content_type)

            # Check that the correct number of zones match
            self.assertEqual(correct_result, len(response.json['zones']))

    def test_invalid_zones_filter(self):
        invalid_url = '/zones?id=155477ef-e6c5-4b94-984d-8fc68c0c1a14'
        self._assert_exception(
            'bad_request', 400, self.client.get, invalid_url)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_utils import timeutils
from sqlalchemy import (Integer, String, Unicode, DateTime,
                        ForeignKeyConstraint, UniqueConstraint)
from sqlalchemy.schema import Table, Column, MetaData

from designate import utils
from designate.sqlalchemy.types import UUID

meta = MetaData()

pool_nameservers = Table('pool_nameservers', meta,
    Column('id', UUID, default=utils.generate_uuid, primary_key=True),
    Column('version', Integer(), default=1, nullable=False),
    Column('created_at', DateTime, default=lambda: timeutils.utcnow()),
    Column('updated_at', DateTime, onupdate=lambda: timeutils.utcnow()),

    Column('pool_id', UUID(), nullable=False),
    Column('host', String(255), nullable=False),
    Column('port', Integer(), nullable=False),

    ForeignKeyConstraint(['pool_id'], ['pools.id'], ondelete='CASCADE'),
    UniqueConstraint('pool_id', 'host', 'port', name='unique_pool_host_port'),

    mysql_engine='InnoDB',
    mysql_charset='utf8',
)

pool_targets = Table('pool_targets', meta,
    Column('id', UUID, default=utils.generate_uuid, primary_key=True),
    Column('version', Integer(), default=1, nullable=False),
    Column('created_at', DateTime, default=lambda: timeutils.utcnow()),
    Column('updated_at', DateTime, onupdate=lambda: timeutils.utcnow()),

    Column('pool_id', UUID(), nullable=False),
    Column('type', String(50), nullable=False),
    Column('tsigkey_id', UUID(), nullable=True),
    Column('description', Unicode(160), nullable=True),

    ForeignKeyConstraint(['pool_id'], ['pools.id'], ondelete='CASCADE'),

    mysql_engine='InnoDB',
    mysql_charset='utf8',
)

pool_target_masters = Table('pool_target_masters', meta,
    Column('id', UUID, default=utils.generate_uuid, primary_key=True),
    Column('version', Integer(), default=1, nullable=False),
    Column('created_at', DateTime, default=lambda: timeutils.utcnow()),
    Column('updated_at', DateTime, onupdate=lambda: timeutils.utcnow()),

    Column('pool_target_id', UUID(), nullable=False),
    Column('host', String(255), nullable=False),
    Column('port', Integer(), nullable=False),

    ForeignKeyConstraint(['pool_target_id'], ['pool_targets.id'],
                         ondelete='CASCADE'),
    UniqueConstraint('pool_target_id', 'host', 'port',
                     name='unique_pool_target_host_port'),

    mysql_engine='InnoDB',
    mysql_charset='utf8',
)

pool_target_options = Table('pool_target_options', meta,
    Column('id', UUID, default=utils.generate_uuid, primary_key=True),
    Column('version', Integer(), default=1, nullable=False),
    Column('created_at', DateTime, default=lambda: timeutils.utcnow()),
    Column('updated_at', DateTime, onupdate=lambda: timeutils.utcnow()),

    Column('pool_target_id', UUID(), nullable=False),
    Column('key', String(255), nullable=False),
    Column('value', String(255), nullable=False),

    ForeignKeyConstraint(['pool_target_id'], ['pool_targets.id'],
                         ondelete='CASCADE'),
    UniqueConstraint('pool_target_id', 'key', name='unique_pool_target_key'),

    mysql_engine='InnoDB',
    mysql_charset='utf8',
)

pool_also_notifies = Table('pool_also_notifies', meta,
    Column('id', UUID, default=utils.generate_uuid, primary_key=True),
    Column('version', Integer(), default=1, nullable=False),
    Column('created_at', DateTime, default=lambda: timeutils.utcnow()),
    Column('updated_at', DateTime, onupdate=lambda: timeutils.utcnow()),

    Column('pool_id', UUID(), nullable=False),
    Column('host', String(255), nullable=False),
    Column('port', Integer(), nullable=False),

    ForeignKeyConstraint(['pool_id'], ['pools.id'], ondelete='CASCADE'),
    UniqueConstraint('pool_id', 'host', 'port',
                     name='unique_pool_also_notifies_pool0host0port'),

    mysql_engine='InnoDB',
    mysql_charset='utf8',
)


def upgrade(migrate_engine):
    meta.bind = migrate_engine

    # Load the pool_attributes_table table schema for relations
    Table('pools', meta, autoload=True)

    pool_nameservers.create(checkfirst=True)
    pool_targets.create(checkfirst=True)
    pool_target_options.create(checkfirst=True)
    pool_target_masters.create(checkfirst=True)
    pool_also_notifies.create(checkfirst=True)



# Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
# Author: Federico Ceratto <federico.ceratto@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest.mock import call
from unittest import mock

from oslo_concurrency import processutils

import designate.tests
from designate import exceptions
from designate.backend.agent_backend import impl_knot2
from designate.tests.unit.agent import backends


class Knot2AgentBackendTestCase(designate.tests.TestCase):
    def setUp(self):
        super(Knot2AgentBackendTestCase, self).setUp()

        self.backend = impl_knot2.Knot2Backend('foo')
        self.backend._execute_knotc = mock.Mock()

    def test_start_backend(self):
        self.backend.start()

    def test_stop_backend(self):
        self.backend.stop()

    def test_create_zone(self):
        zone = backends.create_dnspy_zone('example.org')

        self.backend.create_zone(zone)

        self.backend._execute_knotc.assert_has_calls([
            call('conf-begin'),
            call('conf-set', 'zone[example.org]',
                 expected_error='duplicate identifier'),
            call('conf-commit'),
            call('zone-refresh', 'example.org')
        ])

    def test_create_zone_already_there(self):
        self.backend._execute_knotc.return_value = 'duplicate identifier'

        zone = backends.create_dnspy_zone('example.org')

        self.backend.create_zone(zone)

        self.backend._execute_knotc.assert_has_calls([
            call('conf-begin'),
            call('conf-set', 'zone[example.org]',
                 expected_error='duplicate identifier'),
            call('conf-commit'),
            call('zone-refresh', 'example.org')
        ])

    def test_start_minidns_to_knot_axfr(self):
        self.backend._start_minidns_to_knot_axfr('foo')

        self.backend._execute_knotc.assert_called_with('zone-refresh', 'foo')

    @mock.patch('oslo_concurrency.lockutils.lock')
    def test_modify_zone(self, mock_lock):
        self.backend._modify_zone('blah', 'bar')

        self.assertEqual(3, self.backend._execute_knotc.call_count)

        self.backend._execute_knotc.assert_called_with('conf-commit')

    @mock.patch('oslo_concurrency.lockutils.lock')
    def test_modify_zone_exception(self, mock_lock):
        # Raise an exception during the second call to _execute_knotc
        self.backend._execute_knotc.side_effect = [None, exceptions.Backend,
                                                   None]
        self.assertRaises(
            exceptions.Backend,
            self.backend._modify_zone, 'blah', 'bar'
        )

        self.assertEqual(3, self.backend._execute_knotc.call_count)

        self.backend._execute_knotc.assert_has_calls([
            call('conf-begin'),
            call('blah', 'bar'),
            call('conf-abort'),
        ])

    @mock.patch('designate.backend.agent_backend.impl_knot2.execute')
    def test_find_zone_serial(self, mock_execute):
        result = (
            '[example.com.] type: slave | serial: 20 | next-event: idle | '
            'auto-dnssec: disabled]'
        )
        mock_execute.return_value = result, ''

        serial = self.backend.find_zone_serial('example.com')

        self.assertEqual(20, serial)

    @mock.patch('designate.backend.agent_backend.impl_knot2.execute')
    def test_find_zone_serial_zone_not_found(self, mock_execute):
        mock_execute.side_effect = processutils.ProcessExecutionError(
            'error: [example.com.] (no such zone found)'
        )

        serial = self.backend.find_zone_serial('example.com')

        self.assertIsNone(serial)

    @mock.patch('designate.backend.agent_backend.impl_knot2.execute')
    def test_find_zone_serial_unexpected_output(self, mock_execute):
        mock_execute.return_value = 'bogus output', ''

        self.assertRaises(
            exceptions.Backend,
            self.backend.find_zone_serial, 'example.com'
        )

    @mock.patch('designate.backend.agent_backend.impl_knot2.execute')
    def test_find_zone_serial_error(self, mock_execute):
        mock_execute.side_effect = processutils.ProcessExecutionError('blah')

        self.assertRaises(
            exceptions.Backend,
            self.backend.find_zone_serial, 'example.com'
        )

    def test_update_zone(self):
        zone = backends.create_dnspy_zone('example.org')

        self.backend.update_zone(zone)

        self.backend._execute_knotc.assert_called_once_with(
            'zone-refresh', 'example.org'
        )

    def test_delete_zone(self):
        self.backend.delete_zone('example.org')

        self.backend._execute_knotc.assert_has_calls([
            call('conf-begin'),
            call('conf-unset', 'zone[example.org]',
                 expected_error='invalid identifier'),
            call('conf-commit'),
        ])

    def test_delete_zone_already_gone(self):
        self.backend._execute_knotc.return_value = 'duplicate identifier'

        self.backend.delete_zone('example.org')

        self.backend._execute_knotc.assert_has_calls([
            call('conf-begin'),
            call('conf-unset', 'zone[example.org]',
                 expected_error='invalid identifier'),
            call('conf-commit'),
        ])


class Knot2AgentExecuteTestCase(designate.tests.TestCase):
    def setUp(self):
        super(Knot2AgentExecuteTestCase, self).setUp()

        self.backend = impl_knot2.Knot2Backend('foo')

    def test_init(self):
        self.assertEqual('knotc', self.backend._knotc_cmd_name)

    @mock.patch('designate.backend.agent_backend.impl_knot2.execute')
    def test_execute_knotc_ok(self, mock_execute):
        mock_execute.return_value = ('OK', '')

        self.backend._execute_knotc('a1', 'a2')

        mock_execute.assert_called_with('knotc', 'a1', 'a2')

        self.assertEqual(1, mock_execute.call_count)

    @mock.patch('designate.backend.agent_backend.impl_knot2.execute')
    def test_execute_knotc_expected_error(self, mock_execute):
        mock_execute.return_value = ('xyz', '')

        self.backend._execute_knotc('a1', 'a2', expected_error='xyz')

        mock_execute.assert_called_once_with('knotc', 'a1', 'a2')

    @mock.patch('designate.backend.agent_backend.impl_knot2.execute')
    def test_execute_knotc_expected_output(self, mock_execute):
        mock_execute.return_value = ('xyz', '')

        self.backend._execute_knotc('a1', 'a2', expected_output='xyz')

        mock_execute.assert_called_once_with('knotc', 'a1', 'a2')

    @mock.patch('designate.backend.agent_backend.impl_knot2.execute')
    def test_execute_knotc_with_error(self, mock_execute):
        mock_execute.return_value = ('xyz', '')

        self.assertRaises(
            exceptions.Backend,
            self.backend._execute_knotc, 'a1', 'a2'
        )

        mock_execute.assert_called_once_with('knotc', 'a1', 'a2')

    @mock.patch('designate.backend.agent_backend.impl_knot2.execute')
    def test_execute_knotc_raising_exception(self, mock_execute):
        mock_execute.side_effect = processutils.ProcessExecutionError

        self.assertRaises(
            exceptions.Backend,
            self.backend._execute_knotc, 'a1', 'a2'
        )

        mock_execute.assert_called_once_with('knotc', 'a1', 'a2')



#
# Copyright 2014 Red Hat, Inc.
# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import math

from oslo_concurrency import lockutils
from oslo_log import log
import tenacity
import tooz.coordination

import designate.conf
from designate import utils

CONF = designate.conf.CONF
LOG = log.getLogger(__name__)


def _retry_if_tooz_error(exception):
    """Return True if we should retry, False otherwise"""
    return isinstance(exception, tooz.coordination.ToozError)


class Coordination(object):
    def __init__(self, name, tg, grouping_enabled=False):
        # NOTE(eandersson): Workaround until tooz handles the conversion.
        if not isinstance(name, bytes):
            name = name.encode('ascii')
        self.name = name
        self.tg = tg
        self.coordination_id = None
        self._grouping_enabled = grouping_enabled
        self._coordinator = None
        self._started = False

    @property
    def coordinator(self):
        return self._coordinator

    @property
    def started(self):
        return self._started

    def get_lock(self, name):
        if self._coordinator:
            # NOTE(eandersson): Workaround until tooz handles the conversion.
            if not isinstance(name, bytes):
                name = name.encode('ascii')
            return self._coordinator.get_lock(name)
        return lockutils.lock(name)

    def start(self):
        self.coordination_id = ":".join([CONF.host, utils.generate_uuid()])
        self._started = False

        backend_url = CONF.coordination.backend_url
        if backend_url is None:
            LOG.warning('No coordination backend configured, distributed '
                        'coordination functionality will be disabled. '
                        'Please configure a coordination backend.')
            return

        self._coordinator = tooz.coordination.get_coordinator(
            backend_url, self.coordination_id.encode()
        )
        while not self._coordinator.is_started:
            self._coordinator.start(start_heart=True)

        self._started = True

        if self._grouping_enabled:
            self._enable_grouping()

    def stop(self):
        if self._coordinator is None:
            return

        try:
            if self._grouping_enabled:
                self._disable_grouping()
            self._coordinator.stop()
            self._coordinator = None
        finally:
            self._started = False

    def _coordinator_run_watchers(self):
        if not self._started:
            return
        self._coordinator.run_watchers()

    def _create_group(self):
        try:
            create_group_req = self._coordinator.create_group(
                self.name
            )
            create_group_req.get()
        except tooz.coordination.GroupAlreadyExist:
            pass
        join_group_req = self._coordinator.join_group(self.name)
        join_group_req.get()

    def _disable_grouping(self):
        try:
            leave_group_req = self._coordinator.leave_group(self.name)
            leave_group_req.get()
        except tooz.coordination.GroupNotCreated:
            pass

    def _enable_grouping(self):
        self._create_group()
        self.tg.add_timer(
            CONF.coordination.run_watchers_interval,
            self._coordinator_run_watchers
        )


class Partitioner(object):
    def __init__(self, coordinator, group_id, my_id, partitions):
        self._coordinator = coordinator
        self._group_id = group_id
        self._my_id = my_id
        self._partitions = partitions

        self._started = False
        self._my_partitions = None
        self._callbacks = []

    def _warn_no_backend(self):
        LOG.warning('No coordination backend configured, assuming we are '
                    'the only worker. Please configure a coordination '
                    'backend')

    @tenacity.retry(stop=tenacity.stop_after_attempt(5),
                    wait=tenacity.wait_random(max=2),
                    retry=tenacity.retry_if_exception(_retry_if_tooz_error),
                    reraise=True)
    def _get_members(self, group_id):
        get_members_req = self._coordinator.get_members(group_id)
        try:
            return get_members_req.get()

        except tooz.coordination.GroupNotCreated:
            LOG.error('Attempting to partition over a non-existent group: %s',
                      self._group_id)

            raise
        except tooz.coordination.ToozError:
            LOG.error('Error getting group membership info from coordination '
                      'backend.')
            raise

    def _on_group_change(self, event):
        LOG.debug('Received member change %s', event)
        members, self._my_partitions = self._update_partitions()

        self._run_callbacks(members, event)

    def _partition(self, members, me, partitions):
        member_count = len(members)
        partition_count = len(partitions)
        partition_size = int(
            math.ceil(float(partition_count) / float(member_count)))

        my_index = members.index(me)
        my_start = partition_size * my_index
        my_end = my_start + partition_size

        return partitions[my_start:my_end]

    def _run_callbacks(self, members, event):
        for cb in self._callbacks:
            cb(self.my_partitions, members, event)

    def _update_partitions(self):
        # Recalculate partitions - we need to sort the list of members
        # alphabetically so that it's the same order across all nodes.
        members = sorted(list(self._get_members(self._group_id)))
        partitions = self._partition(
            members, self._my_id, self._partitions)
        return members, partitions

    @property
    def my_partitions(self):
        return self._my_partitions

    def start(self):
        """Allow the partitioner to start timers after the coordinator has
        gotten it's connections up.
        """
        LOG.debug('Starting partitioner')
        if self._coordinator:
            self._coordinator.watch_join_group(
                self._group_id, self._on_group_change)
            self._coordinator.watch_leave_group(
                self._group_id, self._on_group_change)

            # We need to get our partitions now. Events doesn't help in this
            # case since they require state change in the group that we wont
            # get when starting initially
            self._my_partitions = self._update_partitions()[1]
        else:
            self._my_partitions = self._partitions
            self._run_callbacks(None, None)

        self._started = True

    def watch_partition_change(self, callback):
        LOG.debug('Watching for change %s', self._group_id)
        self._callbacks.append(callback)
        if self._started:
            if not self._coordinator:
                self._warn_no_backend()
            callback(self._my_partitions, None, None)

    def unwatch_partition_change(self, callback):
        self._callbacks.remove(callback)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add Unique constraint on ('pool_id', 'hostname') in the pool_ns_records
table Bug #1517389"""

import sys

from migrate.changeset.constraint import UniqueConstraint
from oslo_log import log as logging
from sqlalchemy.schema import MetaData, Table
from sqlalchemy import exc
from sqlalchemy.engine.reflection import Inspector

LOG = logging.getLogger()

meta = MetaData()

CONSTRAINT_NAME = "unique_ns_name"

explanation = """
You need to manually remove duplicate entries from the database.

The error message was:
%s
"""


def upgrade(migrate_engine):
    meta.bind = migrate_engine

    pool_ns_records_table = Table('pool_ns_records', meta, autoload=True)

    # Only apply it if it's not there (It's been backported to L)
    insp = Inspector.from_engine(migrate_engine)
    unique_constraints = insp.get_unique_constraints('pool_ns_records')
    unique_constraint_names = [i['name'] for i in unique_constraints]

    if CONSTRAINT_NAME not in unique_constraint_names:
        # We define the constraint here if not it shows in the list above.
        constraint = UniqueConstraint('pool_id', 'hostname',
                                      name=CONSTRAINT_NAME,
                                      table=pool_ns_records_table)
        try:
            constraint.create()
        except exc.IntegrityError as e:
            LOG.error(explanation, e)
            # Use sys.exit so we don't blow up with a huge trace
            sys.exit(1)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.mport threading
from unittest import mock

import designate.tests
import designate.rpc
from designate.sink import service
from designate.tests import fixtures


class TestSinkService(designate.tests.TestCase):
    def setUp(self):
        super(TestSinkService, self).setUp()
        self.stdlog = fixtures.StandardLogging()
        self.useFixture(self.stdlog)

        self.CONF.set_override('enabled_notification_handlers', ['fake'],
                               'service:sink')

        self.service = service.Service()

    @mock.patch.object(designate.rpc, 'get_notification_listener')
    def test_service_start(self, mock_notification_listener):
        self.service.start()

        self.assertTrue(mock_notification_listener.called)

    def test_service_stop(self):
        self.service.stop()

        self.assertIn('Stopping sink service', self.stdlog.logger.output)

    def test_service_name(self):
        self.assertEqual('sink', self.service.service_name)



# Copyright (c) 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import binascii
from unittest import mock
from unittest.mock import patch

import dns
import dns.message
import dns.query
import dns.exception

from designate.tests.test_mdns import MdnsTestCase
from designate.mdns import notify
from designate import objects


class MdnsNotifyTest(MdnsTestCase):

    test_zone = {
        'name': 'example.com.',
        'email': 'example@example.com',
        'serial': 100,
    }

    def setUp(self):
        super(MdnsNotifyTest, self).setUp()
        self.nameserver = objects.PoolNameserver.from_dict({
            'id': 'f278782a-07dc-4502-9177-b5d85c5f7c7e',
            'host': '127.0.0.1',
            'port': 65255
        })
        self.mock_tg = mock.Mock()
        self.notify = notify.NotifyEndpoint(self.mock_tg)

    def test_send_notify_message(self):
        # id 10001
        # opcode NOTIFY
        # rcode NOERROR
        # flags QR AA
        # ;QUESTION
        # example.com. IN SOA
        # ;ANSWER
        # ;AUTHORITY
        # ;ADDITIONAL
        expected_notify_response = ("2711a4000001000000000000076578616d706c650"
                                    "3636f6d0000060001")
        context = self.get_context()
        with patch.object(dns.query, 'udp', return_value=dns.message.from_wire(
                binascii.a2b_hex(expected_notify_response))):
            response, retry = self.notify.notify_zone_changed(
                context, objects.Zone.from_dict(self.test_zone),
                self.nameserver.host, self.nameserver.port, 0, 0, 2, 0)
            self.assertEqual(response, dns.message.from_wire(
                binascii.a2b_hex(expected_notify_response)))
            self.assertEqual(retry, 1)

    def test_send_notify_message_non_auth(self):
        # id 10001
        # opcode NOTIFY
        # rcode NOTAUTH
        # flags QR
        # ;QUESTION
        # example.com. IN SOA
        # ;ANSWER
        # ;AUTHORITY
        # ;ADDITIONAL
        non_auth_notify_response = ("2711a4090001000000000000076578616d706c650"
                                    "3636f6d0000060001")
        context = self.get_context()
        with patch.object(dns.query, 'udp', return_value=dns.message.from_wire(
                binascii.a2b_hex(non_auth_notify_response))):
            response, retry = self.notify.notify_zone_changed(
                context, objects.Zone.from_dict(self.test_zone),
                self.nameserver.host, self.nameserver.port, 0, 0, 2, 0)
            self.assertIsNone(response)
            self.assertEqual(retry, 1)

    @patch.object(dns.query, 'udp', side_effect=dns.exception.Timeout)
    def test_send_notify_message_timeout(self, _):
        context = self.get_context()
        response, retry = self.notify.notify_zone_changed(
            context, objects.Zone.from_dict(self.test_zone),
            self.nameserver.host, self.nameserver.port, 0, 0, 2, 0)
        self.assertIsNone(response)
        self.assertEqual(retry, 2)

    @patch.object(dns.query, 'udp', side_effect=dns.query.BadResponse)
    def test_send_notify_message_bad_response(self, _):
        context = self.get_context()
        response, retry = self.notify.notify_zone_changed(
            context, objects.Zone.from_dict(self.test_zone),
            self.nameserver.host, self.nameserver.port, 0, 0, 2, 0)
        self.assertIsNone(response)
        self.assertEqual(retry, 1)

    def test_poll_for_serial_number(self):
        # id 10001
        # opcode QUERY
        # rcode NOERROR
        # flags QR AA
        # ;QUESTION
        # example.com. IN SOA
        # ;ANSWER
        # example.com. 3600 IN SOA example-ns.com. admin.example.com. 100 3600
        #  600 86400 3600
        # ;AUTHORITY
        # ;ADDITIONAL
        poll_response = ("271184000001000100000000076578616d706c6503636f6d0000"
                         "060001c00c0006000100000e1000290a6578616d706c652d6e73"
                         "c0140561646d696ec00c0000006400000e100000025800015180"
                         "00000e10")
        context = self.get_context()
        with patch.object(dns.query, 'udp', return_value=dns.message.from_wire(
                binascii.a2b_hex(poll_response))):
            status, serial, retries = self.notify.get_serial_number(
                context, objects.Zone.from_dict(self.test_zone),
                self.nameserver.host, self.nameserver.port, 0, 0, 2, 0)
            self.assertEqual(status, 'SUCCESS')
            self.assertEqual(serial, self.test_zone['serial'])
            self.assertEqual(retries, 2)

    def test_poll_for_serial_number_lower_serial(self):
        # id 10001
        # opcode QUERY
        # rcode NOERROR
        # flags QR AA
        # ;QUESTION
        # example.com. IN SOA
        # ;ANSWER
        # example.com. 3600 IN SOA example-ns.com. admin.example.com. 99 3600
        #  600 86400 3600
        # ;AUTHORITY
        # ;ADDITIONAL
        poll_response = ("271184000001000100000000076578616d706c6503636f6d0000"
                         "060001c00c0006000100000e1000290a6578616d706c652d6e73"
                         "c0140561646d696ec00c0000006300000e100000025800015180"
                         "00000e10")
        context = self.get_context()
        with patch.object(dns.query, 'udp', return_value=dns.message.from_wire(
                binascii.a2b_hex(poll_response))):
            status, serial, retries = self.notify.get_serial_number(
                context, objects.Zone.from_dict(self.test_zone),
                self.nameserver.host, self.nameserver.port, 0, 0, 2, 0)
            self.assertEqual(status, 'ERROR')
            self.assertEqual(serial, 99)
            self.assertEqual(retries, 0)

    def test_poll_for_serial_number_higher_serial(self):
        # id 10001
        # opcode QUERY
        # rcode NOERROR
        # flags QR AA
        # ;QUESTION
        # example.com. IN SOA
        # ;ANSWER
        # example.com. 3600 IN SOA example-ns.com. admin.example.com. 101 3600
        #  600 86400 3600
        # ;AUTHORITY
        # ;ADDITIONAL
        poll_response = ("271184000001000100000000076578616d706c6503636f6d0000"
                         "060001c00c0006000100000e1000290a6578616d706c652d6e73"
                         "c0140561646d696ec00c0000006500000e100000025800015180"
                         "00000e10")
        context = self.get_context()
        with patch.object(dns.query, 'udp', return_value=dns.message.from_wire(
                binascii.a2b_hex(poll_response))):
            status, serial, retries = self.notify.get_serial_number(
                context, objects.Zone.from_dict(self.test_zone),
                self.nameserver.host, self.nameserver.port, 0, 0, 2, 0)
            self.assertEqual(status, 'SUCCESS')
            self.assertEqual(serial, 101)
            self.assertEqual(retries, 2)

    @patch.object(dns.query, 'udp', side_effect=dns.exception.Timeout)
    def test_poll_for_serial_number_timeout(self, _):
        context = self.get_context()
        status, serial, retries = self.notify.get_serial_number(
            context, objects.Zone.from_dict(self.test_zone),
            self.nameserver.host, self.nameserver.port, 0, 0, 2, 0)
        self.assertEqual(status, 'ERROR')
        self.assertIsNone(serial)
        self.assertEqual(retries, 0)

    @patch('dns.query.udp', side_effect=dns.exception.Timeout)
    @patch('dns.query.tcp', side_effect=dns.exception.Timeout)
    def test_send_dns_message_all_tcp(self, tcp, udp):
        self.config(
            all_tcp=True,
            group='service:mdns'
        )
        context = self.get_context()
        test_zone = objects.Zone.from_dict(self.test_zone)
        status, serial, retries = self.notify.get_serial_number(
            context, test_zone, self.nameserver.host,
            self.nameserver.port, 0, 0, 2, 0)
        response, retry = self.notify.notify_zone_changed(
            context, test_zone, self.nameserver.host,
            self.nameserver.port, 0, 0, 2, 0)
        assert not udp.called
        assert tcp.called



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_policy import policy

from designate.common.policies import base

rules = [
    policy.DocumentedRuleDefault(
        name="create_zone",
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Create Zone",
        operations=[
            {
                'path': '/v2/zones',
                'method': 'POST'
            }
        ]
    ),
    policy.RuleDefault(
        name="get_zones",
        check_str=base.RULE_ADMIN_OR_OWNER
    ),
    policy.DocumentedRuleDefault(
        name="get_zone",
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Get Zone",
        operations=[
            {
                'path': '/v2/zones/{zone_id}',
                'method': 'GET'
            }, {
                'path': '/v2/zones/{zone_id}',
                'method': 'PATCH'
            }, {
                'path': '/v2/zones/{zone_id}/recordsets/{recordset_id}',
                'method': 'PUT'
            }
        ]
    ),
    policy.RuleDefault(
        name="get_zone_servers",
        check_str=base.RULE_ADMIN_OR_OWNER
    ),
    policy.DocumentedRuleDefault(
        name="find_zones",
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="List existing zones",
        operations=[
            {
                'path': '/v2/zones',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="update_zone",
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Update Zone",
        operations=[
            {
                'path': '/v2/zones/{zone_id}',
                'method': 'PATCH'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="delete_zone",
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Delete Zone",
        operations=[
            {
                'path': '/v2/zones/{zone_id}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="xfr_zone",
        check_str=base.RULE_ADMIN_OR_OWNER,
        description="Manually Trigger an Update of a Secondary Zone",
        operations=[
            {
                'path': '/v2/zones/{zone_id}/tasks/xfr',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="abandon_zone",
        check_str=base.RULE_ADMIN,
        description="Abandon Zone",
        operations=[
            {
                'path': '/v2/zones/{zone_id}/tasks/abandon',
                'method': 'POST'
            }
        ]
    ),
    policy.RuleDefault(
        name="count_zones",
        check_str=base.RULE_ADMIN_OR_OWNER
    ),
    policy.RuleDefault(
        name="count_zones_pending_notify",
        check_str=base.RULE_ADMIN_OR_OWNER
    ),
    policy.RuleDefault(
        name="purge_zones",
        check_str=base.RULE_ADMIN
    ),
    policy.RuleDefault(
        name="touch_zone",
        check_str=base.RULE_ADMIN_OR_OWNER
    )
]


def list_rules():
    return rules



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

from designate import quota
from designate import tests

LOG = logging.getLogger(__name__)


class StorageQuotaTest(tests.TestCase):
    def setUp(self):
        super(StorageQuotaTest, self).setUp()
        self.config(quota_driver='storage')
        self.quota = quota.get_quota()

    def test_set_quota_create(self):
        context = self.get_admin_context()
        context.all_tenants = True

        quota = self.quota.set_quota(context, 'tenant_id', 'zones', 1500)

        self.assertEqual({'zones': 1500}, quota)

        # Drop into the storage layer directly to ensure the quota was created
        # successfully
        criterion = {
            'tenant_id': 'tenant_id',
            'resource': 'zones'
        }

        quota = self.quota.storage.find_quota(context, criterion)

        self.assertEqual('tenant_id', quota['tenant_id'])
        self.assertEqual('zones', quota['resource'])
        self.assertEqual(1500, quota['hard_limit'])

    def test_set_quota_update(self):
        context = self.get_admin_context()
        context.all_tenants = True

        # First up, Create the quota
        self.quota.set_quota(context, 'tenant_id', 'zones', 1500)

        # Next, update the quota
        self.quota.set_quota(context, 'tenant_id', 'zones', 1234)

        # Drop into the storage layer directly to ensure the quota was updated
        # successfully
        criterion = {
            'tenant_id': 'tenant_id',
            'resource': 'zones'
        }

        quota = self.quota.storage.find_quota(context, criterion)

        self.assertEqual('tenant_id', quota['tenant_id'])
        self.assertEqual('zones', quota['resource'])
        self.assertEqual(1234, quota['hard_limit'])

    def test_reset_quotas(self):
        context = self.get_admin_context()
        context.all_tenants = True

        # First up, Create a zones quota
        self.quota.set_quota(context, 'tenant_id', 'zones', 1500)

        # Then, Create a zone_records quota
        self.quota.set_quota(context, 'tenant_id', 'zone_records', 800)

        # Now, Reset the tenants quota
        self.quota.reset_quotas(context, 'tenant_id')

        # Drop into the storage layer directly to ensure the tenant has no
        # specific quotas registed.
        criterion = {
            'tenant_id': 'tenant_id'
        }

        quotas = self.quota.storage.find_quotas(context, criterion)
        self.assertEqual(0, len(quotas))



# Copyright (c) 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from designate.objects import base
from designate.objects import fields


@base.DesignateRegistry.register
class TsigKey(base.DictObjectMixin, base.PersistentObjectMixin,
              base.DesignateObject):
    def __init__(self, *args, **kwargs):
        super(TsigKey, self).__init__(*args, **kwargs)

    fields = {
        'name': fields.StringFields(nullable=False, maxLength=160),
        'algorithm': fields.EnumField(nullable=False,
            valid_values=[
                'hmac-md5',
                'hmac-sha1',
                'hmac-sha224',
                'hmac-sha256',
                'hmac-sha384',
                'hmac-sha512'
            ]
        ),
        'secret': fields.StringFields(maxLength=160),
        'scope': fields.EnumField(nullable=False,
            valid_values=['POOL', 'ZONE']
        ),
        'resource_id': fields.UUIDFields(nullable=False)
    }

    STRING_KEYS = [
        'id', 'name', 'algorithm', 'scope', 'resource_id'
    ]


@base.DesignateRegistry.register
class TsigKeyList(base.ListObjectMixin, base.DesignateObject):
    LIST_ITEM_TYPE = TsigKey

    fields = {
        'objects': fields.ListOfObjectsField('TsigKey'),
    }



# coding=utf-8
# COPYRIGHT 2014 Rackspace
#
# Author: Tim Simmons <tim.simmons@rackspace.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg

from designate.tests.test_api.test_admin import AdminApiTestCase

cfg.CONF.import_opt('enabled_extensions_admin', 'designate.api.admin',
                    group='service:api')


class AdminApiQuotasTest(AdminApiTestCase):
    def setUp(self):
        self.config(enabled_extensions_admin=['quotas'], group='service:api')
        super(AdminApiQuotasTest, self).setUp()

    def test_get_quotas(self):
        self.policy({'get_quotas': '@'})
        context = self.get_admin_context()

        response = self.client.get('/quotas/%s' % context.project_id,
                                   headers={'X-Test-Tenant-Id':
                                   context.project_id})

        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        self.assertIn('quota', response.json)
        self.assertIn('zones', response.json['quota'])
        self.assertIn('api_export_size', response.json['quota'])
        self.assertIn('zone_records', response.json['quota'])
        self.assertIn('zone_recordsets', response.json['quota'])
        self.assertIn('recordset_records', response.json['quota'])

        max_zones = response.json['quota']['zones']
        max_zone_records = response.json['quota']['zone_records']
        self.assertEqual(cfg.CONF.quota_zones, max_zones)
        self.assertEqual(cfg.CONF.quota_zone_records, max_zone_records)

    def test_patch_quotas(self):
        self.policy({'set_quotas': '@'})
        context = self.get_context(project_id='a', is_admin=True)

        response = self.client.get('/quotas/%s' % 'a',
                                   headers={'X-Test-Tenant-Id':
                                   context.project_id})

        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        self.assertIn('quota', response.json)
        self.assertIn('zones', response.json['quota'])
        current_count = response.json['quota']['zones']

        body = {'quota': {"zones": 1337}}
        response = self.client.patch_json('/quotas/%s' % 'a', body,
                                          status=200,
                                          headers={'X-Test-Tenant-Id':
                                          context.project_id})
        self.assertEqual(200, response.status_int)

        response = self.client.get('/quotas/%s' % 'a',
                                   headers={'X-Test-Tenant-Id':
                                   context.project_id})

        new_count = response.json['quota']['zones']

        self.assertNotEqual(current_count, new_count)

    def test_patch_quotas_validation(self):
        self.policy({'set_quotas': '@'})
        context = self.get_context(project_id='a', is_admin=True)
        url = '/quotas/%s' % 'a'

        # Test a negative number for zones
        body = {'quota': {"zones": -1337}}

        # Ensure it fails with a 400
        self._assert_exception(
            'invalid_object', 400, self.client.patch_json,
            url, body,
            headers={'X-Test-Tenant-Id': context.project_id}
        )

        # Test a number > maximum (2147483647) for zones
        body = {'quota': {"zones": 2147483648}}

        # Ensure it fails with a 400
        self._assert_exception(
            'invalid_object', 400, self.client.patch_json,
            url, body,
            headers={'X-Test-Tenant-Id': context.project_id}
        )

    def test_reset_quotas(self):
        self.policy({'reset_quotas': '@'})
        context = self.get_context(project_id='a', is_admin=True)

        response = self.client.get('/quotas/%s' % 'a',
                                   headers={'X-Test-Tenant-Id':
                                   context.project_id})

        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        self.assertIn('quota', response.json)
        self.assertIn('zones', response.json['quota'])
        current_count = response.json['quota']['zones']

        body = {'quota': {"zones": 1337}}
        response = self.client.patch_json('/quotas/%s' % 'a', body,
                                          status=200,
                                          headers={'X-Test-Tenant-Id':
                                          context.project_id})
        self.assertEqual(200, response.status_int)

        response = self.client.get('/quotas/%s' % 'a',
                                   headers={'X-Test-Tenant-Id':
                                   context.project_id})

        new_count = response.json['quota']['zones']

        self.assertNotEqual(current_count, new_count)

        response = self.client.delete('/quotas/%s' % 'a',
                                   headers={'X-Test-Tenant-Id':
                                   context.project_id}, status=204)
        response = self.client.get('/quotas/%s' % 'a',
                                   headers={'X-Test-Tenant-Id':
                                   context.project_id})

        newest_count = response.json['quota']['zones']
        self.assertNotEqual(new_count, newest_count)
        self.assertEqual(current_count, newest_count)



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Author: Alin Balutoiu <abalutoiu@cloudbasesolutions.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from os_win import constants
from os_win import exceptions as os_win_exc
from os_win import utilsfactory

import designate.tests
from designate import exceptions
from designate.backend.agent_backend import impl_msdns
from designate.tests.unit.agent import backends


class MSDNSAgentBackendTestCase(designate.tests.TestCase):
    @mock.patch.object(utilsfactory, 'get_dnsutils')
    def setUp(self, mock_get_dnsutils):
        super(MSDNSAgentBackendTestCase, self).setUp()
        self.zone_name = 'example.com'

        self.CONF.set_override('masters', ['127.0.0.1:5354'], 'service:agent')

        self.backend = impl_msdns.MSDNSBackend('foo')
        self.backend._dnsutils = mock.MagicMock()

    def test_start_backend(self):
        self.backend.start()

    def test_stop_backend(self):
        self.backend.stop()

    def test_init(self):
        self.assertEqual(['127.0.0.1'], self.backend._masters)

    @mock.patch.object(utilsfactory, 'get_dnsutils')
    def test_init_no_masters(self, mock_get_dnsutils):
        self.CONF.set_override('masters', [], 'service:agent')

        self.assertRaisesRegex(
            exceptions.Backend,
            'Missing agent AXFR masters',
            impl_msdns.MSDNSBackend, 'foo'
        )

    def test_find_zone_serial(self):
        serial = self.backend.find_zone_serial(self.zone_name)

        expected_serial = self.backend._dnsutils.get_zone_serial.return_value
        self.assertEqual(expected_serial, serial)

        self.backend._dnsutils.get_zone_serial.assert_called_once_with(
            self.zone_name
        )

    def test_find_zone_serial_error(self):
        self.backend._dnsutils.get_zone_serial.side_effect = (
            os_win_exc.DNSZoneNotFound(zone_name=self.zone_name))

        serial = self.backend.find_zone_serial(self.zone_name)

        self.assertIsNone(serial)
        self.backend._dnsutils.get_zone_serial.assert_called_once_with(
            self.zone_name
        )

    def test_create_zone(self):
        zone = backends.create_dnspy_zone(self.zone_name)

        self.backend.create_zone(zone)

        self.backend._dnsutils.zone_create.assert_called_once_with(
            zone_name=self.zone_name,
            zone_type=constants.DNS_ZONE_TYPE_SECONDARY,
            ds_integrated=False,
            ip_addrs=self.backend._masters
        )

    def test_create_zone_already_existing_diff(self):
        zone = backends.create_dnspy_zone(self.zone_name)
        self.backend._dnsutils.zone_create.side_effect = (
            os_win_exc.DNSZoneAlreadyExists(zone_name=self.zone_name))

        self.assertRaises(
            os_win_exc.DNSZoneAlreadyExists,
            self.backend.create_zone, zone
        )

        self.backend._dnsutils.get_zone_properties.assert_called_once_with(
            self.zone_name
        )

    def test_create_zone_already_existing_identical(self):
        zone = backends.create_dnspy_zone(self.zone_name)
        self.backend._dnsutils.zone_create.side_effect = (
            os_win_exc.DNSZoneAlreadyExists(zone_name=self.zone_name)
        )

        mock_zone_properties = {
            'zone_type': constants.DNS_ZONE_TYPE_SECONDARY,
            'ds_integrated': False,
            'master_servers': self.backend._masters
        }
        self.backend._dnsutils.get_zone_properties.return_value = (
            mock_zone_properties
        )

        self.backend.create_zone(zone)

        self.backend._dnsutils.get_zone_properties.assert_called_once_with(
            self.zone_name
        )

    def test_update_zone(self):
        zone = backends.create_dnspy_zone(self.zone_name)

        self.backend.update_zone(zone)

        self.backend._dnsutils.zone_update.assert_called_once_with(
            self.zone_name
        )

    def test_delete_zone(self):
        self.backend.delete_zone(self.zone_name)

        self.backend._dnsutils.zone_delete.assert_called_once_with(
            self.zone_name
        )



# -*- coding: utf-8 -*-
# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from oslo_log import log as logging

from designate.tests import TestCase
from designate.notification_handler.nova import NovaFixedHandler
from designate.tests.test_notification_handler import \
    NotificationHandlerMixin

LOG = logging.getLogger(__name__)


class NovaFixedHandlerTest(TestCase, NotificationHandlerMixin):
    def setUp(self):
        super(NovaFixedHandlerTest, self).setUp()

        zone = self.create_zone()
        self.zone_id = zone['id']
        self.config(zone_id=zone['id'], group='handler:nova_fixed')
        self.config(formatv4=['%(host)s.%(zone)s',
                              '%(host)s.foo.%(zone)s'],
                    formatv6=['%(host)s.%(zone)s',
                              '%(host)s.foo.%(zone)s'],
                    group='handler:nova_fixed')

        self.plugin = NovaFixedHandler()

    def test_instance_create_end(self):
        event_type = 'compute.instance.create.end'
        fixture = self.get_notification_fixture('nova', event_type)

        self.assertIn(event_type, self.plugin.get_event_types())

        criterion = {'zone_id': self.zone_id}

        # Ensure we start with 2 records
        records = self.central_service.find_records(self.admin_context,
                                                    criterion)
        # Should only be SOA and NS records
        self.assertEqual(2, len(records))

        self.plugin.process_notification(
            self.admin_context.to_dict(), event_type, fixture['payload'])

        # Ensure we now have exactly 1 more record
        records = self.central_service.find_records(self.admin_context,
                                                    criterion)

        self.assertEqual(4, len(records))

    def test_instance_create_end_utf8(self):
        self.config(formatv4=['%(display_name)s.%(zone)s'],
                    formatv6=['%(display_name)s.%(zone)s'],
                    group='handler:nova_fixed')

        event_type = 'compute.instance.create.end'
        fixture = self.get_notification_fixture('nova', event_type)

        # Set the instance display_name to a string containing UTF8.
        fixture['payload']['display_name'] = u'Test↟Instance'

        self.assertIn(event_type, self.plugin.get_event_types())

        criterion = {'zone_id': self.zone_id}

        # Ensure we start with 2 records
        recordsets = self.central_service.find_recordsets(
            self.admin_context, criterion)

        # Should only be SOA and NS recordsets
        self.assertEqual(2, len(recordsets))

        self.plugin.process_notification(
            self.admin_context.to_dict(), event_type, fixture['payload'])

        # Ensure we now have exactly 1 more recordset
        recordsets = self.central_service.find_recordsets(
            self.admin_context, criterion)

        self.assertEqual(3, len(recordsets))

        # Ensure the created record was correctly converted per IDN rules.
        criterion['type'] = 'A'
        recordsets = self.central_service.find_recordsets(
            self.admin_context, criterion)

        self.assertEqual('xn--testinstance-q83g.example.com.',
                         recordsets[0].name)

    def test_instance_delete_start(self):
        # Prepare for the test
        start_event_type = 'compute.instance.create.end'
        start_fixture = self.get_notification_fixture('nova', start_event_type)

        self.plugin.process_notification(self.admin_context.to_dict(),
                                         start_event_type,
                                         start_fixture['payload'])

        # Now - Onto the real test
        event_type = 'compute.instance.delete.start'
        fixture = self.get_notification_fixture('nova', event_type)

        self.assertIn(event_type, self.plugin.get_event_types())

        criterion = {'zone_id': self.zone_id}

        # Ensure we start with at least 1 record, plus NS and SOA
        records = self.central_service.find_records(self.admin_context,
                                                    criterion)

        self.assertEqual(4, len(records))

        self.plugin.process_notification(
            self.admin_context.to_dict(), event_type, fixture['payload'])

        # Simulate the record having been deleted on the backend
        zone_serial = self.central_service.get_zone(
            self.admin_context, self.zone_id).serial
        self.central_service.update_status(
            self.admin_context, self.zone_id, "SUCCESS", zone_serial)

        # Ensure we now have exactly 0 records, plus NS and SOA
        records = self.central_service.find_records(self.admin_context,
                                                    criterion)

        self.assertEqual(2, len(records))

    def test_label_in_format_v4_v6(self):
        event_type = 'compute.instance.create.end'
        self.config(formatv4=['%(label)s.example.com.'],
                    formatv6=['%(label)s.example.com.'],
                    group='handler:nova_fixed')
        fixture = self.get_notification_fixture('nova', event_type)
        with mock.patch.object(
                self.plugin, '_create_or_update_recordset') as finder:
            with mock.patch.object(self.plugin.central_api,
                                   'create_record'):
                finder.return_value = {'id': 'fakeid'}
                self.plugin.process_notification(
                    self.admin_context.to_dict(),
                    event_type, fixture['payload'])
                finder.assert_called_once_with(
                    mock.ANY, mock.ANY, type='A', zone_id=self.zone_id,
                    name='private.example.com.')

    def test_formatv4(self):
        event_type = 'compute.instance.create.end'
        self.config(formatv4=['%(label)s-v4.example.com.'],
                    group='handler:nova_fixed')
        fixture = self.get_notification_fixture('nova', event_type)
        with mock.patch.object(
                self.plugin, '_create_or_update_recordset') as finder:
            with mock.patch.object(self.plugin.central_api,
                                   'create_record'):
                finder.return_value = {'id': 'fakeid'}
                self.plugin.process_notification(
                    self.admin_context.to_dict(),
                    event_type, fixture['payload'])
                finder.assert_called_once_with(
                    mock.ANY, mock.ANY, type='A', zone_id=self.zone_id,
                    name='private-v4.example.com.')

    def test_formatv6(self):
        event_type = 'compute.instance.create.end'
        self.config(formatv6=['%(label)s-v6.example.com.'],
                    group='handler:nova_fixed')
        fixture = self.get_notification_fixture('nova', event_type)
        with mock.patch.object(
                self.plugin, '_create_or_update_recordset') as finder:
            with mock.patch.object(self.plugin.central_api,
                                   'create_record'):
                finder.return_value = {'id': 'fakeid'}
                self.plugin.process_notification(
                    self.admin_context.to_dict(),
                    event_type, fixture['payload_v6'])
                finder.assert_called_once_with(
                    mock.ANY, mock.ANY, type='AAAA', zone_id=self.zone_id,
                    name='private-v6.example.com.')



# Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import inspect

LOG = {}


def log_rpc_call(func, rpcapi, logger):
    def wrapped(*args, **kwargs):
        logger.debug("Calling designate.%(rpcapi)s.%(function)s() "
                     "over RPC", {'function': func.__name__,
                                  'rpcapi': rpcapi})
        return func(*args, **kwargs)
    return wrapped


LOGGING_BLACKLIST = [
    'get_instance',
    '__init__'
]


def rpc_logging(logger, rpcapi):
    def wrapper(cls):
        CLASS_BLACKLIST = getattr(cls, 'LOGGING_BLACKLIST', [])
        BLACKLIST = CLASS_BLACKLIST + LOGGING_BLACKLIST
        for name, m in inspect.getmembers(cls, inspect.ismethod):
            if name not in BLACKLIST:
                setattr(cls, name, log_rpc_call(m, rpcapi, logger))
        return cls
    return wrapper



# Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from sqlalchemy.schema import MetaData


meta = MetaData()


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine

    if migrate_engine.name != "mysql":
        return

    migrate_engine.execute("SET foreign_key_checks = 0;")
    migrate_engine.execute(
        "ALTER TABLE service_statuses CONVERT TO CHARACTER SET utf8;")
    migrate_engine.execute("SET foreign_key_checks = 1;")
    migrate_engine.execute(
        "ALTER DATABASE %s DEFAULT CHARACTER SET utf8;"
        % migrate_engine.url.database)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# Copied: nova.notifications
import abc

from oslo_config import cfg
from oslo_log import log as logging
import six

import designate.conf
from designate.plugin import DriverPlugin
from designate import objects
from designate import rpc

LOG = logging.getLogger(__name__)
CONF = designate.conf.CONF
NOTIFICATION_PLUGIN = None


def send_api_fault(context, url, status, exception):
    """Send an api.fault notification."""

    if not CONF.notify_api_faults:
        return

    payload = {'url': url, 'exception': str(exception), 'status': status}

    rpc.get_notifier('api').error(context, 'dns.api.fault', payload)


def init_notification_plugin():
    LOG.debug("Loading notification plugin: %s", cfg.CONF.notification_plugin)
    cls = NotificationPlugin.get_driver(cfg.CONF.notification_plugin)

    global NOTIFICATION_PLUGIN
    NOTIFICATION_PLUGIN = cls()


def get_plugin():
    if NOTIFICATION_PLUGIN is None:
        init_notification_plugin()
    return NOTIFICATION_PLUGIN


class NotificationPlugin(DriverPlugin):
    """Base class for Notification Driver implementations"""
    __plugin_type__ = 'notification'
    __plugin_ns__ = 'designate.notification.plugin'

    def __init__(self):
        super(NotificationPlugin, self).__init__()

    @abc.abstractmethod
    def emit(self, notification_type, context, result, *args, **kwargs):
        """Return a payload to emit as part of the notification"""


class Default(NotificationPlugin):
    """Returns the result, as implemented in the base class"""
    __plugin_name__ = 'default'

    def emit(self, notification_type, context, result, *args, **kwargs):
        """Return the result of the function called"""
        return [result]


class Audit(NotificationPlugin):
    """Grabs Zone/Recordset names and RRData changes"""
    __plugin_name__ = 'audit'

    def zone_name(self, arglist, result):
        for arg in arglist + [result]:
            if isinstance(arg, objects.Zone):
                if arg.name is not None:
                    return arg.name
            if hasattr(arg, 'zone_name'):
                if arg.zone_name is not None:
                    return arg.zone_name

        return None

    def zone_id(self, arglist, result):
        for arg in arglist + [result]:
            if isinstance(arg, objects.Zone):
                if arg.id is not None:
                    return arg.id
            if hasattr(arg, 'zone_id'):
                if arg.zone_id is not None:
                    return arg.zone_id

        return None

    def recordset_name(self, arglist, result):
        for arg in arglist + [result]:
            if isinstance(arg, objects.RecordSet):
                if arg.name is not None:
                    return arg.name

        return None

    def recordset_data(self, arglist, result):
        if not isinstance(result, objects.RecordSet):
            return []

        for arg in arglist:
            if isinstance(arg, objects.RecordSet):
                if 'records' not in arg.obj_what_changed():
                    return []
                original_rrs = arg.obj_get_original_value('records')

                old_value = ' '.join(
                    [obj['data'] for obj in original_rrs])

                new_value = ' '.join(
                    [rr.data for rr in result.records])

                if old_value == new_value:
                    return []

                return [{
                    'change': 'records',
                    'old_value': old_value,
                    'new_value': new_value,
                }]

        return []

    def other_data(self, arglist, result):
        changes = []

        for arg in arglist:
            if isinstance(arg, objects.DesignateObject):
                for change in arg.obj_what_changed():
                    if change != 'records':
                        old_value = arg.obj_get_original_value(change)
                        new_value = getattr(arg, change)

                        # Just in case something odd makes it here
                        if any(not isinstance(val,
                                              (int, float, bool,
                                               six.string_types, type(None)))
                               for val in (old_value, new_value)):
                            LOG.warning("Nulling notification values after "
                                        "unexpected values (%s, %s)",
                                        old_value, new_value)
                            old_value, new_value = None, None

                        if old_value == new_value:
                            continue

                        changes.append({
                            'change': change,
                            'old_value': str(old_value),
                            'new_value': str(new_value),
                        })

        return changes

    def blank_event(self):
        return [{
            'change': None,
            'old_value': None,
            'new_value': None,
        }]

    def gather_changes(self, arglist, result, notification_type):
        changes = []

        if 'update' in notification_type:
            changes.extend(self.other_data(arglist, result))
            if notification_type == 'dns.recordset.update':
                changes.extend(self.recordset_data(arglist, result))
        elif 'create' in notification_type:
            if notification_type == 'dns.recordset.create':
                changes.extend(self.recordset_data(arglist, result))
            else:
                changes.extend(self.blank_event())
        else:
            changes.extend(self.blank_event())

        return changes

    def emit(self, notification_type, context, result, *args, **kwargs):
        arglist = []
        for item in args:
            if isinstance(item, tuple) or isinstance(item, list):
                arglist.extend(item)
            if isinstance(item, dict):
                arglist.extend(list(item.values()))

        payloads = []
        for change in self.gather_changes(arglist, result, notification_type):
            payloads.append({
                'zone_name': self.zone_name(arglist, result),
                'zone_id': self.zone_id(arglist, result),
                'recordset_name': self.recordset_name(arglist, result),
                'old_data': change['old_value'],
                'new_data': change['new_value'],
                'changed_field': change['change'],
            })

        return payloads



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from testscenarios import load_tests_apply_scenarios as load_tests  # noqa
import testtools
from oslo_config import cfg
from oslo_log import log as logging

from designate import quota
from designate import tests
from designate import exceptions


LOG = logging.getLogger(__name__)


class QuotaTestCase(tests.TestCase):
    scenarios = [
        ('noop', dict(quota_driver='noop')),
        ('storage', dict(quota_driver='storage'))
    ]

    def setUp(self):
        super(QuotaTestCase, self).setUp()
        self.config(quota_driver=self.quota_driver)
        self.quota = quota.get_quota()

    def test_get_quotas(self):
        context = self.get_admin_context()

        quotas = self.quota.get_quotas(context, 'DefaultQuotaTenant')

        self.assertIsNotNone(quotas)
        self.assertEqual({
            'api_export_size': cfg.CONF.quota_api_export_size,
            'zones': cfg.CONF.quota_zones,
            'zone_recordsets': cfg.CONF.quota_zone_recordsets,
            'zone_records': cfg.CONF.quota_zone_records,
            'recordset_records': cfg.CONF.quota_recordset_records,
        }, quotas)

    def test_limit_check_unknown(self):
        context = self.get_admin_context()

        with testtools.ExpectedException(exceptions.QuotaResourceUnknown):
            self.quota.limit_check(context, 'tenant_id', unknown=0)

        with testtools.ExpectedException(exceptions.QuotaResourceUnknown):
            self.quota.limit_check(context, 'tenant_id', unknown=0, zones=0)

    def test_limit_check_under(self):
        context = self.get_admin_context()

        self.quota.limit_check(context, 'tenant_id', zones=0)
        self.quota.limit_check(context, 'tenant_id', zone_records=0)
        self.quota.limit_check(context, 'tenant_id', zones=0,
                               zone_records=0)

        self.quota.limit_check(context, 'tenant_id',
                               zones=(cfg.CONF.quota_zones - 1))
        self.quota.limit_check(
            context,
            'tenant_id',
            zone_records=(cfg.CONF.quota_zone_records - 1))

    def test_limit_check_at(self):
        context = self.get_admin_context()

        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id',
                                   zones=cfg.CONF.quota_zones)

        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(
                context,
                'tenant_id',
                zone_records=cfg.CONF.quota_zone_records)

    def test_limit_check_unlimited(self):
        context = self.get_admin_context()
        self.quota.get_quotas = mock.Mock()
        ret = {
                'zones': -1,
                'zone_recordsets': -1,
                'zone_records': -1,
                'recordset_records': -1,
                'api_export_size': -1,
        }
        self.quota.get_quotas.return_value = ret
        self.quota.limit_check(context, 'tenant_id', zones=99999)
        self.quota.limit_check(context, 'tenant_id', zone_recordsets=99999)
        self.quota.limit_check(context, 'tenant_id', zone_records=99999)
        self.quota.limit_check(context, 'tenant_id', recordset_records=99999)
        self.quota.limit_check(context, 'tenant_id', api_export_size=99999)

    def test_limit_check_zero(self):
        context = self.get_admin_context()
        self.quota.get_quotas = mock.Mock()
        ret = {
                'zones': 0,
                'zone_recordsets': 0,
                'zone_records': 0,
                'recordset_records': 0,
                'api_export_size': 0,
        }
        self.quota.get_quotas.return_value = ret
        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id', zones=0)
        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id', zone_recordsets=0)
        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id', zone_records=0)
        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id',
                                   recordset_records=0)
        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id', api_export_size=0)

    def test_limit_check_over(self):
        context = self.get_admin_context()

        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id', zones=99999)

        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id', zone_records=99999)

        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id', zones=99999,
                                   zone_records=99999)

        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id', zones=99999,
                                   zone_records=0)

        with testtools.ExpectedException(exceptions.OverQuota):
            self.quota.limit_check(context, 'tenant_id', zones=0,
                                   zone_records=99999)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Author: Federico Ceratto <federico.ceratto@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import oslotest.base
from oslo_config import cfg
from oslo_config import fixture as cfg_fixture

import designate.dnsutils
import designate.rpc
import designate.service
from designate import storage
import designate.utils
from designate.mdns import handler
from designate.mdns import service
from designate.tests import fixtures

CONF = cfg.CONF


class MdnsServiceTest(oslotest.base.BaseTestCase):
    @mock.patch.object(storage, 'get_storage', mock.Mock())
    def setUp(self):
        super(MdnsServiceTest, self).setUp()
        self.stdlog = fixtures.StandardLogging()
        self.useFixture(self.stdlog)

        self.useFixture(cfg_fixture.Config(CONF))

        self.service = service.Service()

    @mock.patch.object(designate.service.DNSService, 'start')
    @mock.patch.object(designate.service.RPCService, 'start')
    def test_service_start(self, mock_rpc_start, mock_dns_start):
        self.service.start()

        self.assertTrue(mock_dns_start.called)
        self.assertTrue(mock_rpc_start.called)

    def test_service_stop(self):
        self.service.dns_service.stop = mock.Mock()

        self.service.stop()

        self.assertTrue(self.service.dns_service.stop.called)

        self.assertIn('Stopping mdns service', self.stdlog.logger.output)

    def test_service_name(self):
        self.assertEqual('mdns', self.service.service_name)

    def test_mdns_rpc_topic(self):
        CONF.set_override('topic', 'test-topic', 'service:mdns')

        self.service = service.Service()

        self.assertEqual('test-topic', self.service.rpc_topic)
        self.assertEqual('mdns', self.service.service_name)

    @mock.patch.object(storage, 'get_storage')
    def test_storage_driver(self, mock_get_driver):
        self.service._storage = None

        mock_driver = mock.MagicMock()
        mock_driver.name = 'noop_driver'
        mock_get_driver.return_value = mock_driver

        self.assertIsInstance(self.service.storage, mock.MagicMock)

        self.assertTrue(mock_get_driver.called)

    @mock.patch.object(handler, 'RequestHandler')
    @mock.patch.object(designate.service.DNSService, 'start')
    @mock.patch.object(designate.utils, 'cache_result')
    def test_dns_application(self, mock_cache_result, mock_dns_start,
                             mock_request_handler):

        app = self.service.dns_application

        self.assertIsInstance(app, designate.dnsutils.DNSMiddleware)



# Copyright 2014 Rackspace Inc.
#
# Author: Eric Larson <eric.larson@rackspace.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import dns
from oslo_config import cfg
from oslo_config import fixture as cfg_fixture
import oslotest.base

from designate import exceptions
from designate import objects
from designate.mdns import handler
from designate.tests import fixtures

CONF = cfg.CONF


class MdnsHandleTest(oslotest.base.BaseTestCase):
    def setUp(self):
        super(MdnsHandleTest, self).setUp()
        self.stdlog = fixtures.StandardLogging()
        self.useFixture(self.stdlog)
        self.useFixture(cfg_fixture.Config(CONF))
        self.context = mock.Mock()
        self.storage = mock.Mock()
        self.tg = mock.Mock()
        self.handler = handler.RequestHandler(self.storage, self.tg)

    @mock.patch.object(dns.resolver.Resolver, 'query')
    def test_notify(self, mock_query):
        self.storage.find_zone.return_value = objects.Zone(
            id='e2bed4dc-9d01-11e4-89d3-123b93f75cba',
            serial=2,
            masters=objects.ZoneMasterList.from_list([
                {'host': '1.0.0.0', 'port': 53},
            ])
        )
        mock_query.return_value = [
            mock.Mock(serial=1)
        ]

        request = dns.message.make_query(
            'www.example.org.', dns.rdatatype.SOA
        )
        request.environ = dict(addr=['1.0.0.0'], context=self.context)

        response = self.handler._handle_notify(request)

        self.assertEqual(dns.rcode.NOERROR, tuple(response)[0].rcode())

        self.assertIn(
            'Scheduling AXFR for e2bed4dc-9d01-11e4-89d3-123b93f75cba '
            'from 1.0.0.0:53',
            self.stdlog.logger.output
        )

    @mock.patch.object(dns.resolver.Resolver, 'query')
    def test_notify_same_serial(self, mock_query):
        self.storage.find_zone.return_value = objects.Zone(
            id='e2bed4dc-9d01-11e4-89d3-123b93f75cba',
            serial=1,
            masters=objects.ZoneMasterList.from_list([
                {'host': '1.0.0.0', 'port': 53},
            ])
        )
        mock_query.return_value = [
            mock.Mock(serial=1)
        ]

        request = dns.message.make_query(
            'www.example.org.', dns.rdatatype.SOA
        )
        request.environ = dict(addr=['1.0.0.0'], context=self.context)

        response = self.handler._handle_notify(request)

        self.assertEqual(dns.rcode.NOERROR, tuple(response)[0].rcode())

        self.assertIn(
            'Serial 1 is the same for master and us for '
            'e2bed4dc-9d01-11e4-89d3-123b93f75cba',
            self.stdlog.logger.output
        )

    def test_notify_no_questions(self):
        request = dns.message.make_query(
            'www.example.org.', dns.rdatatype.SOA
        )
        request.environ = dict(context=self.context)
        request.question = []

        response = self.handler._handle_notify(request)

        self.assertEqual(dns.rcode.FORMERR, tuple(response)[0].rcode())

    def test_notify_zone_not_found(self):
        self.storage.find_zone.side_effect = exceptions.ZoneNotFound

        request = dns.message.make_query(
            'www.example.org.', dns.rdatatype.SOA
        )
        request.environ = dict(context=self.context)

        response = self.handler._handle_notify(request)

        self.assertEqual(dns.rcode.NOTAUTH, tuple(response)[0].rcode())

    def test_notify_no_master_addr(self):
        self.storage.find_zone.return_value = objects.Zone(
            masters=objects.ZoneMasterList.from_list([
                {'host': '1.0.0.0', 'port': 53},
            ])
        )

        request = dns.message.make_query(
            'www.example.org.', dns.rdatatype.SOA
        )
        request.environ = dict(addr=['127.0.0.1', 53], context=self.context)

        response = self.handler._handle_notify(request)

        self.assertEqual(dns.rcode.REFUSED, tuple(response)[0].rcode())

        self.assertIn(
            'NOTIFY for None from non-master server 127.0.0.1, refusing.',
            self.stdlog.logger.output
        )

    def test_axfr_zone_not_found(self):
        self.storage.find_zone.side_effect = exceptions.ZoneNotFound

        request = dns.message.make_query(
            'www.example.org.', dns.rdatatype.AXFR
        )
        request.environ = dict(context=self.context)

        response = tuple(self.handler._handle_axfr(request))

        self.assertEqual(dns.rcode.REFUSED, response[0].rcode())

        self.assertIn(
            'ZoneNotFound while handling axfr request. '
            'Question was www.example.org. IN AXFR',
            self.stdlog.logger.output
        )

    def test_axfr_forbidden(self):
        self.storage.find_zone.side_effect = exceptions.Forbidden

        request = dns.message.make_query(
            'www.example.org.', dns.rdatatype.AXFR
        )
        request.environ = dict(context=self.context)

        response = tuple(self.handler._handle_axfr(request))

        self.assertEqual(dns.rcode.REFUSED, response[0].rcode())

        self.assertIn(
            'Forbidden while handling axfr request. '
            'Question was www.example.org. IN AXFR',
            self.stdlog.logger.output
        )

    def test_get_max_message_size(self):
        CONF.set_override('max_message_size', 32768, 'service:mdns')

        self.assertEqual(
            32768, self.handler._get_max_message_size(had_tsig=False)
        )

    def test_get_max_message_size_larger_than_allowed(self):
        CONF.set_override('max_message_size', 65535 * 2, 'service:mdns')

        self.assertEqual(
            65535, self.handler._get_max_message_size(had_tsig=False)
        )

        self.assertIn(
            'MDNS max message size must not be greater than 65535',
            self.stdlog.logger.output
        )

    def test_get_max_message_with_tsig(self):
        CONF.set_override('max_message_size', 65535, 'service:mdns')

        self.assertEqual(
            65300, self.handler._get_max_message_size(had_tsig=True)
        )


class TestRequestHandlerCall(oslotest.base.BaseTestCase):
    def setUp(self):
        super(TestRequestHandlerCall, self).setUp()
        self.handler = handler.RequestHandler(mock.Mock(), mock.Mock())
        self.handler._central_api = mock.Mock(name='central_api')

        # Use a simple handlers that doesn't require a real request
        self.handler._handle_query_error = mock.Mock(return_value='Error')
        self.handler._handle_axfr = mock.Mock(return_value=['AXFR'])
        self.handler._handle_record_query = mock.Mock(
            return_value=['Record Query'])
        self.handler._handle_notify = mock.Mock(return_value=['Notify'])

    def test_central_api_property(self):
        self.handler._central_api = 'foo'
        self.assertEqual(self.handler.central_api, 'foo')

    def test__call___unhandled_opcodes(self):
        unhandled_codes = [
            dns.opcode.STATUS,
            dns.opcode.IQUERY,
            dns.opcode.UPDATE,
        ]

        request = mock.Mock()
        for code in unhandled_codes:
            request.opcode.return_value = code  # return an error

            self.assertEqual(['Error'], list(self.handler(request)))
            self.handler._handle_query_error.assert_called_with(
                request, dns.rcode.REFUSED
            )

    def test__call__query_error_with_more_than_one_question(self):
        request = mock.Mock()
        request.opcode.return_value = dns.opcode.QUERY
        request.question = [mock.Mock(), mock.Mock()]

        self.assertEqual(['Error'], list(self.handler(request)))
        self.handler._handle_query_error.assert_called_with(
            request, dns.rcode.REFUSED
        )

    def test__call__query_error_with_data_claas_not_in(self):
        request = mock.Mock()
        request.opcode.return_value = dns.opcode.QUERY
        request.question = [mock.Mock(rdclass=dns.rdataclass.ANY)]

        self.assertEqual(['Error'], list(self.handler(request)))
        self.handler._handle_query_error.assert_called_with(
            request, dns.rcode.REFUSED
        )

    def test__call__axfr(self):
        request = mock.Mock()
        request.opcode.return_value = dns.opcode.QUERY
        request.question = [
            mock.Mock(rdclass=dns.rdataclass.IN, rdtype=dns.rdatatype.AXFR)
        ]

        self.assertEqual(['AXFR'], list(self.handler(request)))

    def test__call__ixfr(self):
        request = mock.Mock()
        request.opcode.return_value = dns.opcode.QUERY
        request.question = [
            mock.Mock(rdclass=dns.rdataclass.IN, rdtype=dns.rdatatype.IXFR)
        ]

        self.assertEqual(['AXFR'], list(self.handler(request)))

    def test__call__record_query(self):
        request = mock.Mock()
        request.opcode.return_value = dns.opcode.QUERY
        request.question = [
            mock.Mock(rdclass=dns.rdataclass.IN, rdtype=dns.rdatatype.A)
        ]

        self.assertEqual(['Record Query'], list(self.handler(request)))

    def test__call__notify(self):
        request = mock.Mock()
        request.opcode.return_value = dns.opcode.NOTIFY

        self.assertEqual(['Notify'], list(self.handler(request)))

    def test_convert_to_rrset_no_records(self):
        zone = objects.Zone.from_dict({'ttl': 1234})
        recordset = objects.RecordSet(
            name='www.example.org.',
            type='A',
            records=objects.RecordList(objects=[
            ])
        )

        r_rrset = self.handler._convert_to_rrset(zone, recordset)

        self.assertIsNone(r_rrset)

    def test_convert_to_rrset(self):
        zone = objects.Zone.from_dict({'ttl': 1234})
        recordset = objects.RecordSet(
            name='www.example.org.',
            type='A',
            records=objects.RecordList(objects=[
                objects.Record(data='192.0.2.1'),
                objects.Record(data='192.0.2.2'),
            ])
        )

        r_rrset = self.handler._convert_to_rrset(zone, recordset)

        self.assertEqual(2, len(r_rrset))


class HandleRecordQueryTest(oslotest.base.BaseTestCase):
    def setUp(self):
        super(HandleRecordQueryTest, self).setUp()
        self.context = mock.Mock()
        self.storage = mock.Mock()
        self.handler = handler.RequestHandler(self.storage, mock.Mock())

    def test_handle_record_query_empty_recordlist(self):
        # bug #1550441
        self.storage.find_recordset.return_value = objects.RecordSet(
            name='www.example.org.',
            type='A',
            records=objects.RecordList(objects=[
            ])
        )

        request = dns.message.make_query('www.example.org.', dns.rdatatype.A)
        request.environ = dict(context=self.context)
        response_gen = self.handler._handle_record_query(request)
        for response in response_gen:
            # This was raising an exception due to bug #1550441
            out = response.to_wire(max_size=65535)

            self.assertEqual(33, len(out))

    def test_handle_record_query_zone_not_found(self):
        self.storage.find_recordset.return_value = objects.RecordSet(
            name='www.example.org.',
            type='A',
            records=objects.RecordList(objects=[
                objects.Record(data='192.0.2.2'),
            ])
        )
        self.storage.find_zone.side_effect = exceptions.ZoneNotFound

        request = dns.message.make_query('www.example.org.', dns.rdatatype.A)
        request.environ = dict(context=self.context)
        response = tuple(self.handler._handle_record_query(request))

        self.assertEqual(1, len(response))
        self.assertEqual(dns.rcode.REFUSED, response[0].rcode())

    def test_handle_record_query_forbidden(self):
        self.storage.find_recordset.return_value = objects.RecordSet(
            name='www.example.org.',
            type='A',
            records=objects.RecordList(objects=[
                objects.Record(data='192.0.2.2'),
            ])
        )
        self.storage.find_zone.side_effect = exceptions.Forbidden

        request = dns.message.make_query('www.example.org.', dns.rdatatype.A)
        request.environ = dict(context=self.context)
        response = tuple(self.handler._handle_record_query(request))

        self.assertEqual(1, len(response))
        self.assertEqual(dns.rcode.REFUSED, response[0].rcode())

    def test_handle_record_query_find_recordsed_forbidden(self):
        self.storage.find_recordset.side_effect = exceptions.Forbidden

        request = dns.message.make_query('www.example.org.', dns.rdatatype.A)
        request.environ = dict(context=self.context)
        response = tuple(self.handler._handle_record_query(request))

        self.assertEqual(1, len(response))
        self.assertEqual(dns.rcode.REFUSED, response[0].rcode())

    def test_handle_record_query_find_recordsed_not_found(self):
        self.storage.find_recordset.side_effect = exceptions.NotFound

        request = dns.message.make_query('www.example.org.', dns.rdatatype.A)
        request.environ = dict(context=self.context)
        response = tuple(self.handler._handle_record_query(request))

        self.assertEqual(1, len(response))
        self.assertEqual(dns.rcode.REFUSED, response[0].rcode())



# Copyright 2019 Cloudification GmbH
#
# Author: Sergey Kraynev <contact@cloudification.io>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

import requests

import designate.tests
from designate import exceptions
from designate import objects
from designate.backend import impl_akamai_v2 as akamai
from designate.tests import fixtures


class AkamaiBackendTestCase(designate.tests.TestCase):
    def setUp(self):
        super(AkamaiBackendTestCase, self).setUp()
        self.zone = objects.Zone(
            id='cca7908b-dad4-4c50-adba-fb67d4c556e8',
            name='example.com.',
            email='example@example.com'
        )

        self.target = {
            'id': '4588652b-50e7-46b9-b688-a9bad40a873e',
            'type': 'akamai_v2',
            'masters': [
                {'host': '192.168.1.1', 'port': 53},
                {'host': '192.168.1.2', 'port': 35}
            ],
            'options': [
                {'key': 'host', 'value': '192.168.2.3'},
                {'key': 'port', 'value': '53'},
                {'key': 'akamai_client_secret', 'value': 'client_secret'},
                {'key': 'akamai_host', 'value': 'host_value'},
                {'key': 'akamai_access_token', 'value': 'access_token'},
                {'key': 'akamai_client_token', 'value': 'client_token'},
                {'key': 'akamai_contract_id', 'value': 'G-XYW'},
                {'key': 'akamai_gid', 'value': '777'}
            ],
        }

    def gen_response(self, status_code, reason, json_data=None):
        response = requests.models.Response()
        response.status_code = status_code
        response.reason = reason
        response._content = json.dumps(json_data or {}).encode('utf-8')
        return response

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_create_zone_missed_contract_id(self, mock_post, mock_auth):
        self.target['options'].remove(
            {'key': 'akamai_contract_id', 'value': 'G-XYW'})
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(
            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        with fixtures.random_seed(0):
            self.assertRaisesRegex(
                exceptions.Backend,
                'contractId is required for zone creation',
                backend.create_zone, self.admin_context, self.zone)

        mock_post.assert_not_called()

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_create_zone(self, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(
            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        with fixtures.random_seed(0):
            backend.create_zone(self.admin_context, self.zone)

        project_id = self.admin_context.project_id or self.zone.tenant_id
        mock_post.assert_called_once_with(
            json={
                'comment': 'Created by Designate for Tenant %s' % project_id,
                'masters': ['192.168.1.1', '192.168.1.2'],
                'type': 'secondary', 'zone': u'example.com.'
            },
            params={
                'gid': '777',
                'contractId': 'G-XYW'
            },
            url='https://host_value/config-dns/v2/zones'
        )

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_create_zone_duplicate_zone(self, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(
            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        mock_post.return_value = self.gen_response(409, 'Conflict')

        with fixtures.random_seed(0):
            backend.create_zone(self.admin_context, self.zone)

        project_id = self.admin_context.project_id or self.zone.tenant_id
        mock_post.assert_called_once_with(
            json={
                'comment': 'Created by Designate for Tenant %s' % project_id,
                'masters': ['192.168.1.1', '192.168.1.2'],
                'type': 'secondary', 'zone': u'example.com.'
            },
            params={
                'gid': '777',
                'contractId': 'G-XYW'
            },
            url='https://host_value/config-dns/v2/zones'
        )

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_create_zone_with_tsig_key(self, mock_post, mock_auth):
        self.target['options'].extend([
            {'key': 'tsig_key_name', 'value': 'test_key'},
            {'key': 'tsig_key_algorithm', 'value': 'hmac-sha512'},
            {'key': 'tsig_key_secret', 'value': 'aaaabbbbccc'}
        ])
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(
            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        with fixtures.random_seed(0):
            backend.create_zone(self.admin_context, self.zone)

        project_id = self.admin_context.project_id or self.zone.tenant_id
        mock_post.assert_called_once_with(
            json={
                'comment': 'Created by Designate for Tenant %s' % project_id,
                'masters': ['192.168.1.1', '192.168.1.2'],
                'type': 'secondary',
                'zone': 'example.com.',
                'tsigKey': {
                    'name': 'test_key',
                    'algorithm': 'hmac-sha512',
                    'secret': 'aaaabbbbccc',
                }
            },
            params={
                'gid': '777',
                'contractId': 'G-XYW'
            },
            url='https://host_value/config-dns/v2/zones'
        )

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_create_zone_raise_error(self, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(

            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        json_data = {
            'title': 'Missing parameter',
            'detail': 'Missed A option'
        }
        mock_post.return_value = self.gen_response(
            400, 'Bad Request', json_data)

        with fixtures.random_seed(0):
            self.assertRaisesRegex(
                exceptions.Backend,
                'Zone creation failed due to: Missed A option',
                backend.create_zone, self.admin_context, self.zone)

        project_id = self.admin_context.project_id or self.zone.tenant_id
        mock_post.assert_called_once_with(
            json={
                'comment': 'Created by Designate for Tenant %s' % project_id,
                'masters': ['192.168.1.1', '192.168.1.2'],
                'type': 'secondary', 'zone': 'example.com.'
            },
            params={
                'gid': '777',
                'contractId': 'G-XYW'
            },
            url='https://host_value/config-dns/v2/zones'
        )

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_force_delete_zone(self, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(

            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        mock_post.return_value = self.gen_response(200, 'Success')

        with fixtures.random_seed(0):
            backend.delete_zone(self.admin_context, self.zone)

        mock_post.assert_called_once_with(
            json={
                'zones': ['example.com.']
            },
            params={
                'force': True
            },
            url='https://host_value/config-dns/v2/zones/delete-requests'
        )

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_force_delete_zone_raise_error(self, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(

            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        mock_post.return_value = self.gen_response(
            403, 'Bad Request', {'detail': 'Unexpected error'})

        with fixtures.random_seed(0):
            self.assertRaisesRegex(
                exceptions.Backend,
                'Zone deletion failed due to: Unexpected error',
                backend.delete_zone, self.admin_context, self.zone)

        mock_post.assert_called_once_with(
            json={
                'zones': ['example.com.']
            },
            params={
                'force': True
            },
            url='https://host_value/config-dns/v2/zones/delete-requests'
        )

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_force_delete_zone_raise_error_404(self, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(

            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        mock_post.return_value = self.gen_response(
            404, 'Bad Request', {'detail': 'Unexpected error'})

        with fixtures.random_seed(0):
            backend.delete_zone(self.admin_context, self.zone)

        mock_post.assert_called_once_with(
            json={
                'zones': ['example.com.']
            },
            params={
                'force': True
            },
            url='https://host_value/config-dns/v2/zones/delete-requests'
        )

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    @mock.patch.object(akamai.requests.Session, 'get')
    def test_soft_delete_zone(self, mock_get, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(

            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        mock_post.side_effect = [
            # emulate, when Force=True is forbidden
            self.gen_response(403, 'Forbidden'),
            # emulate request, when Force=False
            self.gen_response(200, 'Success', {'requestId': 'nice_id'}),
        ]

        # emulate max 9 failed attempts and 1 success
        mock_get.side_effect = 9 * [
            self.gen_response(200, 'Success', {'isComplete': False})
        ] + [
            self.gen_response(200, 'Success', {'isComplete': True})
        ]

        with fixtures.random_seed(0), \
                mock.patch.object(akamai.time, 'sleep') as mock_sleep:
            mock_sleep.return_value = None
            backend.delete_zone(self.admin_context, self.zone)

        self.assertEqual(10, mock_sleep.call_count)

        url = 'https://host_value/config-dns/v2/zones/delete-requests/nice_id'
        mock_get.assert_has_calls(9 * [mock.call(url=url)])

        mock_post.assert_has_calls([
            mock.call(
                json={'zones': ['example.com.']},
                params={'force': True},
                url='https://host_value/config-dns/v2/zones/delete-requests'
            ),
            mock.call(
                json={'zones': ['example.com.']},
                params={'force': False},
                url='https://host_value/config-dns/v2/zones/delete-requests'
            )
        ])

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    @mock.patch.object(akamai.requests.Session, 'get')
    def test_soft_delete_zone_failed_after_10_attempts(
            self, mock_get, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(

            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        mock_post.side_effect = [
            # emulate, when Force=True is forbidden
            self.gen_response(403, 'Forbidden'),
            # emulate request, when Force=False
            self.gen_response(200, 'Success', {'requestId': 'nice_id'}),
        ]

        # emulate max 10 failed attempts
        mock_get.side_effect = 10 * [
            self.gen_response(200, 'Success', {'isComplete': False})
        ]

        with fixtures.random_seed(0), \
                mock.patch.object(akamai.time, 'sleep') as mock_sleep:
            mock_sleep.return_value = None
            self.assertRaisesRegex(
                exceptions.Backend,
                'Zone was not deleted after 10 attempts',
                backend.delete_zone, self.admin_context, self.zone)

        self.assertEqual(10, mock_sleep.call_count)

        url = 'https://host_value/config-dns/v2/zones/delete-requests/nice_id'
        mock_get.assert_has_calls(10 * [mock.call(url=url)])

        mock_post.assert_has_calls([
            mock.call(
                json={'zones': ['example.com.']},
                params={'force': True},
                url='https://host_value/config-dns/v2/zones/delete-requests'
            ),
            mock.call(
                json={'zones': ['example.com.']},
                params={'force': False},
                url='https://host_value/config-dns/v2/zones/delete-requests'
            )
        ])

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_soft_delete_zone_raise_error(self, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(

            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        mock_post.side_effect = [
            # emulate, when Force=True is forbidden
            self.gen_response(403, 'Forbidden'),
            # emulate request, when Force=False
            self.gen_response(409, 'Conflict', {'detail': 'Intenal Error'})
        ]

        with fixtures.random_seed(0):
            self.assertRaisesRegex(
                exceptions.Backend,
                'Zone deletion failed due to: Intenal Error',
                backend.delete_zone, self.admin_context, self.zone)

        mock_post.assert_has_calls([
            mock.call(
                json={'zones': [u'example.com.']},
                params={'force': True},
                url='https://host_value/config-dns/v2/zones/delete-requests'
            ),
            mock.call(
                json={'zones': [u'example.com.']},
                params={'force': False},
                url='https://host_value/config-dns/v2/zones/delete-requests'
            )
        ])

    @mock.patch.object(akamai, 'edgegrid')
    @mock.patch.object(akamai.requests.Session, 'post')
    def test_soft_delete_zone_missed_request_id(self, mock_post, mock_auth):
        backend = akamai.AkamaiBackend(
            objects.PoolTarget.from_dict(self.target)
        )
        mock_auth.EdgeGridAuth.assert_called_once_with(

            access_token='access_token',
            client_secret='client_secret',
            client_token='client_token'
        )

        mock_post.side_effect = [
            # emulate, when Force=True is forbidden
            self.gen_response(403, 'Forbidden'),
            # emulate request, when Force=False
            self.gen_response(200, 'Success')
        ]

        with fixtures.random_seed(0):
            self.assertRaisesRegex(
                exceptions.Backend,
                'Zone deletion failed due to: requestId missed in response',
                backend.delete_zone, self.admin_context, self.zone)

        mock_post.assert_has_calls([
            mock.call(
                json={'zones': [u'example.com.']},
                params={'force': True},
                url='https://host_value/config-dns/v2/zones/delete-requests'
            ),
            mock.call(
                json={'zones': [u'example.com.']},
                params={'force': False},
                url='https://host_value/config-dns/v2/zones/delete-requests'
            )
        ])



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg
from oslo_db import options

STORAGE_GROUP = cfg.OptGroup(
    name='storage:sqlalchemy',
    title="Configuration for SQLAlchemy Storage"
)


def register_opts(conf):
    conf.register_group(STORAGE_GROUP)
    conf.register_opts(options.database_opts, group=STORAGE_GROUP)


def list_opts():
    return {
        STORAGE_GROUP: options.database_opts,
    }



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import datetime
from unittest.mock import Mock

import oslotest.base
from oslo_log import log as logging
from oslo_utils import timeutils

from designate import objects
from designate.objects import adapters
from designate.objects import base

LOG = logging.getLogger(__name__)


class DesignateTestAdapter(adapters.DesignateAdapter):
    ADAPTER_OBJECT = objects.DesignateObject
    ADAPTER_FORMAT = 'TEST_API'

    MODIFICATIONS = {
        'fields': {},
        'options': {}
    }


@base.DesignateRegistry.register
class DesignateTestPersistentObject(objects.DesignateObject,
                                    objects.base.PersistentObjectMixin):
    pass


class DesignateDateTimeAdaptor(adapters.DesignateAdapter):
    ADAPTER_OBJECT = DesignateTestPersistentObject
    ADAPTER_FORMAT = 'TEST_API'

    MODIFICATIONS = {
        'fields': {
            "id": {},
            "created_at": {},
            "updated_at": {},
        },
        'options': {}
    }


class DesignateAdapterTest(oslotest.base.BaseTestCase):
    def test_get_object_adapter(self):
        adapters.DesignateAdapter.get_object_adapter(
            'TEST_API', objects.DesignateObject())

    def test_object_render(self):
        adapters.DesignateAdapter.render('TEST_API', objects.DesignateObject())

    def test_datetime_format(self):
        test_obj = DesignateTestPersistentObject()
        test_obj.created_at = timeutils.utcnow()

        test_dict = adapters.DesignateAdapter.render('TEST_API', test_obj)

        datetime.datetime.strptime(
            test_dict['created_at'], '%Y-%m-%dT%H:%M:%S.%f'
        )


class RecordSetAPIv2AdapterTest(oslotest.base.BaseTestCase):
    def test_get_path(self):
        request = Mock()
        request.path = '/v2/recordsets'
        recordset = Mock()
        recordset.zone_id = 'a-b-c-d'
        expected_path = '/v2/zones/a-b-c-d/recordsets'

        path = adapters.RecordSetAPIv2Adapter._get_path(request, recordset)
        self.assertEqual(expected_path, path)



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import uuid

from sqlalchemy.types import TypeDecorator, CHAR
from sqlalchemy.dialects.postgresql import UUID as pgUUID


class UUID(TypeDecorator):
    """Platform-independent UUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    Copied verbatim from SQLAlchemy documentation.
    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(pgUUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return str(uuid.UUID(value))



# Copyright 2015 Dyn Inc.
#
# Author: Yasha Bubnov <ybubnov@dyn.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from oslo_config import cfg

import designate.tests
from designate import exceptions
from designate import tests
from designate import utils
from designate.backend.agent_backend import impl_denominator
from designate.tests.unit.agent import backends


class DenominatorAgentBackendTestCase(designate.tests.TestCase):
    def setUp(self):
        super(DenominatorAgentBackendTestCase, self).setUp()

        self.CONF.set_override('listen', ['0.0.0.0:0'], 'service:agent')

        self.backend = impl_denominator.DenominatorBackend('foo')

    def test_start_backend(self):
        self.backend.start()

    def test_stop_backend(self):
        self.backend.stop()

    @mock.patch('designate.utils.execute', return_value=(
            'example.org SOA 86400 ns1.designate.com. '
            'hostmaster@example.org. 475 3600 600 604800 1800', None))
    def test_find_zone_serial(self, mock_execute):
        serial = self.backend.find_zone_serial('example.org.')

        # Ensure returned right serial number
        self.assertEqual(475, serial)

        # Ensure called "denominator zone add"
        self.assertIn('record', mock_execute.call_args[0])
        self.assertIn('get', mock_execute.call_args[0])

    @mock.patch('designate.utils.execute', return_value=('', None))
    def test_find_zone_serial_fail(self, mock_execute):
        serial = self.backend.find_zone_serial('example.org.')
        self.assertIsNone(serial)

    @mock.patch('designate.utils.execute', return_value=(None, None))
    def test_create_zone(self, mock_execute):
        zone = backends.create_dnspy_zone('example.org.')
        self.backend.create_zone(zone)

        # Ensure denominator called for each record (except SOA)
        # plus one to update zone data
        self.assertEqual(
            mock_execute.call_count, len(list(zone.iterate_rdatas()))
        )

    @mock.patch('designate.utils.execute')
    def test_update_zone(self, mock_execute):
        # Output from 'designate record list' command
        records = ('example.org SOA 86400 ns1.designate.com. '
                   'hostmaster@example.org. 475 3600 600 604800 1800\n'
                   'example.org NS 86400 ns1.designator.net.\n'
                   'example.org NS 86400 ns2.designator.net.\n'
                   'example.org MX 86400 10 mx1.designator.net.')

        # That should force update_zone to delete A and AAAA records
        # from the zone and create a new MX record.
        mock_execute.return_value = (records, None)

        zone = backends.create_dnspy_zone('example.org.')
        self.backend.update_zone(zone)

        # Ensure denominator called to:
        # *update zone info
        # *fetch list of zone records
        # *delete one MX record
        # *replace one NS record
        # *create 0 records
        # total: 4 calls

        self.assertEqual(4, mock_execute.call_count)

        methods = ['update_zone',
                   'get_records',
                   'create_record', 'update_record', 'delete_record']
        for method in methods:
            setattr(self.backend.denominator, method, mock.Mock(
                return_value=records))

        self.backend.update_zone(zone)
        self.assertEqual(1, self.backend.denominator.update_zone.call_count)
        self.assertEqual(1, self.backend.denominator.get_records.call_count)
        self.assertEqual(0, self.backend.denominator.create_record.call_count)
        self.assertEqual(1, self.backend.denominator.update_record.call_count)
        self.assertEqual(1, self.backend.denominator.delete_record.call_count)

    @mock.patch('designate.utils.execute', return_value=(None, None))
    def test_delete_zone(self, mock_execute):
        self.backend.delete_zone('example.org.')

        # Ensure called 'denominator zone delete'
        self.assertEqual(1, mock_execute.call_count)
        self.assertIn('zone', mock_execute.call_args[0])
        self.assertIn('delete', mock_execute.call_args[0])


class DenominatorAgentBaseTestCase(tests.TestCase):
    def setUp(self):
        super(DenominatorAgentBaseTestCase, self).setUp()

        self.backend = impl_denominator.Denominator(
            cfg.CONF['backend:agent:denominator']
        )

    def test_base(self):
        self.assertEqual(
            ['denominator', '-q', '-n', 'fake', '-C', '/etc/denominator.conf'],

            self.backend._base()
        )

    def test_base_without_config_file(self):
        self.CONF.set_override(
            'config_file', '', 'backend:agent:denominator'
        )

        self.assertEqual(
            ['denominator', '-q', '-n', 'fake'],
            self.backend._base()
        )

    @mock.patch('designate.utils.execute')
    def test_execute(self, mock_execute):
        mock_execute.return_value = ('stdout', None,)

        self.assertEqual(
            'stdout',
            self.backend._execute(
                ['record', '-z', 'example.org', 'add'], {'name': 'example.org'}
            )
        )

        mock_execute.assert_called_once_with(
            'denominator', '-q', '-n', 'fake', '-C', '/etc/denominator.conf',
            'record', '-z', 'example.org', 'add', '--name', 'example.org'
        )

    @mock.patch('designate.utils.execute')
    def test_execute_raises(self, mock_execute):
        mock_execute.side_effect = utils.processutils.ProcessExecutionError()

        self.assertRaises(
            exceptions.DesignateException,
            self.backend._execute, ['record', '-z', 'example.org', 'add'], {}
        )
